=begin
TIG (c) 2010 - 2013
  All Rights Reserved.
  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
  WARRANTIES,INCLUDING,WITHOUT LIMITATION,THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
[NB:definition and transformation methods are Copyright 2008, AlexM]
###########################################################################
  ImageAnimator.rb
###########################################################################
Needs 'image_class.rb' [comes in the same folder]
###########################################################################
Adds new Tool 'ImageAnimator':

Under Menu > View > ImageAnimator... [submenu]
OR
View Toolbars > ImageAnimator

The Tools include the following:
  "Set Folder for Selected Image"
This sets the 'folder' of sequential 'image-files' that will be used by the 
Image when its animation is run.  The 'folder' chosen should contain images 
with sequential name/numbering as they will load in that order - e.g. as 
they are exported from a GIF or movie file - image001.png, image002.png etc 
They should also all be the same size and filetype[recommended].  PNG files 
can offer transparency so are prefered for 'cutouts' of people etc.  Any 
set 'folder' are remembered across sessions with the SKP.  If an Image has 
a 'folder' set that no longer exists then it is auto-reset to 'nil' and it 
will need redefining.

  "Set Folder from Selected Image to 'Nil'"
This sets the selected Image(s) 'folder' to 'nil' thereby effectively 
removing its animation abilities.  Should an Image has a 'folder' set that 
no longer exists then it is auto-reset to 'nil' and it will need redefining.

  "Set Timer for Selected Images"
This sets the timer [in seconds] for the selected Image(s).  It defaults to 
0.0 - this value means that the Image only animates as the active Scene tab 
changes.  Any value > 0.0 sets the frame-rate in seconds - 0.1 to 0.2 are 
often suitable values - when the Images are animated their animation 
effects continue as you work/pan/zoom or make screen-captures etc.
The shortest workable 'timer' is 0.01 seconds.  Having several animations 
running simultaneously will slow down all animations frame-changes so if 
this situation is expected then reduce the frame rate of all animated 
Images approrpiately.

  "Set FaceMe for Selected Images"
This sets the selected Image(s) 'faceme' behavior.  The choices are:
'None' - the Image does not face the camera [default]
'Bottom-Left' - the Image is placed into a special Group - 'FaceMeImage' - 
which faces the camera, pivoting around the Z_AXIS anchored at its 
bottom-left corner.
'Bottom-Center' - the Image is placed into a special Group - 'FaceMeImage' - 
which faces the camera, pivoting around the Z_AXIS anchored at its 
bottom-center point.
'Bottom-Right' - the Image is placed into a special Group - 'FaceMeImage' - 
which faces the camera, pivoting around the Z_AXIS anchored at its 
bottom-right corner.
A 'FaceMeImage' Group is treated as if it were an Image by the 
ImageAnimator tools.  If later on you reset its 'faceme' behavior to 'None' 
then the Group is exploded and the Image reverts to a plain Image.

  "Animate Selected Images"
This starts the animation effects on the selected Image(s), using the 
'folder' and 'timer' that each one has set as attributes.  Depending on its 
'timer' each Image will animate as the Scene tab changes or continuously as 
you work doing other things.  The first frame in each Image's animation is 
the first image-file in its 'folder'.

  "Stop Selected Images Animation"
This stops the animation effects on the currently selected Image(s).  
Each 'stopped' Image uses its currently loaded image-file.

  "Animate All Images"
This starts the animation effects on ALL suitable Image(s) - i.e those with 
a 'folder' set; using the 'timer' that each one has set as an attribute.  
Depending on its 'timer' each Image will animate as the Scene tab changes 
or continuously as you work doing other things.  The first frame in each 
Image's animation is the first image-file in its 'folder'.

  "Stop All Image Animation"
This stops the animation effects on ALL Image(s).  
Each 'stopped' Image uses its currently loaded image-file.

Menu only items:--------------

  "Make_Unique"
This processes the Selected Image(s) with 'make_unique', i.e. if there is 
more that one Instance of an Image in the Model then the Selected Image is 
given another Definition 0 it is a lone Instance so only it 'animates'.

  "Clone"
If you manually 'Copy' an Image it becomes another Instance of the same 
Image Definition and the Instances will 'animate' together.  You could use 
'Make_Unique' on the Copy to prevent this, alternatively use this 'Clone' 
function - it makes an exact 'copy' of the Selected Image over the original.
However, it is a 'clone' and is 'unique' having a separate Definition, so it 
will 'animate' as a lone Image.

###########################################################################
Donations:
  By Paypal.com to info @ revitrev.org
###########################################################################
Version:
1.0 20101210 First issue.
1.1 20101212 New Menu items added -' Make_Unique' and 'Clone'.
1.2 20101213 Typo corrected in new menu / toolbar code.
1.3 20101216 Typo in menu corrected - again!
1.4 20110105 Clone menu typo fixed.
1.5 20130827 Temp folder resolved.
###########################################################################
=end
###
require('sketchup.rb')
require(File.join(File.dirname(__FILE__), 'image_class.rb'))
###

class ImageAnimator

	def initialize(option=0)
        @option=option
        @msg="ImageAnimator: "
        Sketchup.status_text=@msg
        @model=Sketchup.active_model
        @ss=@model.selection
        @defs=@model.definitions
        case @option
          when 1
            self.set_folder()
          when 2
            self.start_animation()
          when 3
            self.start_animation_all()
          when 4
            self.stop_animation()
          when 5
            self.stop_animation_all()
          when 6
            self.set_timer()
          when 7
            self.set_faceme()
          when 8
            self.no_folder()
          when 9
            self.make_unique()
          when 10
            self.clone()
          else ### no option set
            puts "ImageAnimator Error."
            return nil
        end#case
    end
    
    def set_folder()
        image=nil
        @ss.each{|e|
            if e.class==Sketchup::Image
                image=e
                break
            end#if
            if e.class==Sketchup::Group && e.get_attribute("Animation", "facemeID", nil)
                img=nil
                e.entities.each{|ee|
                    img=ee if ee.class==Sketchup::Image
                }
                if img
                  image=img
                  break
                end#if
            end#if
        }
        if ! image
            UI.messagebox("ImageAnimator: Set Folder:\n\nYou must have an Image Selected.")
            return nil
        end#if
        ### select folder dialog
        path=@model.path
        if ! path || path==""
            dir=Dir.pwd
        else
            dir=File.dirname(path)
        end#if
				###
				path = SKMtools.encoder(path)
				###
        folder=UI.openpanel("Image-files' Folder - Select any Image inside it + 'Open'", dir, ".*")
        ###
        return nil if ! folder
				###
				folder = SKMtools.encoder(folder)
				###
        folder=File.dirname(folder) ### strip off file name
        @model.start_operation("Set Folder")
        image.folder=folder
        @model.commit_operation
        return folder
    end
    
    def no_folder()
        images=[]
        @ss.each{|e|
            images << e if e.class==Sketchup::Image
            if e.class==Sketchup::Group && e.get_attribute("Animation", "facemeID", nil)
                img=nil
                e.entities.each{|ee|
                    img=ee if ee.class==Sketchup::Image
                }
                images << img if img
            end#if
        }
        if ! images[0]
            UI.messagebox("ImageAnimator: Remove Folder:\n\nYou must have at least one Image Selected.")
            return 0
        end#if
        @model.start_operation("Set Folder = 'Nil'")
        images.each{|img|img.folder=nil}
        @model.commit_operation
        return images.length
    end
    
    def start_animation()
        images=[]
        @ss.each{|e|
            if e.class==Sketchup::Image && e.get_attribute("Animation", "folder", nil) && File.exist?(e.get_attribute("Animation", "folder", ""))
                images << e
            end#if
            if e.class==Sketchup::Group && e.get_attribute("Animation", "facemeID", nil)
                img=nil
                e.entities.each{|ee|
                    img=ee if ee.class==Sketchup::Image
                }
                images << img if img
            end#if
        }
        if ! images[0]
            UI.messagebox("ImageAnimator: Animate Selected:\n\nYou must have one or more suitable Images Selected.\n [Images' Image-files Folders must have been preset]")
            return nil
        end#if
        @model.start_operation("Start Animation")
        images.each{|image| image.start_animation() }
        @model.commit_operation
    end
    
    def set_timer()
        images=[]
        @ss.each{|e|
            if e.class==Sketchup::Image && e.get_attribute("Animation", "folder", nil) && File.exist?(e.get_attribute("Animation", "folder", ""))
                images << e
            end#if
            if e.class==Sketchup::Group && e.get_attribute("Animation", "facemeID", nil)
                img=nil
                e.entities.each{|ee|
                    img=ee if ee.class==Sketchup::Image
                }
                images << img if img
            end#if
        }
        if ! images[0]
            UI.messagebox("ImageAnimator: Set Timer for Selected:\n\nYou must have one or more suitable Images Selected.\n")
            return nil
        end#if
        results=UI.inputbox(["Timer [secs]: "], [0.0] ,"ImageAnimator Timer")
        return false if ! results
        timer=results[0]
        timer=nil if timer && timer <= 0.0
        #timer=0.1 if timer && timer < 0.1
        @model.start_operation("Set Timer")
        images.each{|image| image.timer=timer }
        @model.commit_operation
        return timer
    end
    
    def set_faceme()
        images=[]
        @ss.each{|e|
            if e.class==Sketchup::Image
                images << e
            end#if
            if e.class==Sketchup::Group && e.get_attribute("Animation", "facemeID", nil)
                img=nil
                e.entities.each{|ee|
                    img=ee if ee.class==Sketchup::Image
                }
                images << img if img
            end#if
        }
        if ! images[0]
            UI.messagebox("ImageAnimator: Set FaceMe Behavior for Selected:\n\nYou must have one or more suitable Images Selected.\n")
            return nil
        end#if
        results=UI.inputbox(["FaceMe Behavior: "], ["None"], ["None|Bottom-Left|Bottom-Center|Bottom-Right"],"ImageAnimator FaceMe")
        return false if ! results
        type=results[0]
        case type
          when "None"
            faceme=nil
          when "Bottom-Left"
            faceme=1
          when "Bottom-Center"
            faceme=2
          when "Bottom-Right"
            faceme=3
          else
            faceme=nil
        end#case
        @model.start_operation("Set Faceme")
        images.each{|image| image.faceme=faceme }
        @model.commit_operation
        return type
    end
    
    def start_animation_all()
        @model.start_operation("Start Animation All")
        @defs.each{|d|
            if d.image?
                d.instances[0].start_animation_all()
                break
            end#if
        }
        @model.commit_operation
    end
    
    def stop_animation()
        images=[]
        @ss.each{|e|
            if e.class==Sketchup::Image && e.get_attribute("Animation", "folder", nil) && File.exist?(e.get_attribute("Animation", "folder", ""))
                images << e
            end#if
            if e.class==Sketchup::Group && e.get_attribute("Animation", "facemeID", nil)
                img=nil
                e.entities.each{|ee|
                    img=ee if ee.class==Sketchup::Image
                }
                images << img if img
            end#if
        }
        @model.start_operation("Stop Animation")
        images.each{|image| image.stop_animation() }
        @model.commit_operation
    end
    
    def stop_animation_all()
        @model.start_operation("Stop All Animations")
        @defs.each{|d|
            if d.image?
                d.instances[0].stop_animation_all()
                break
            end#if
        }
        @model.commit_operation
    end
    
    def make_unique()
        images=[]
        @ss.each{|e|
            if e.class==Sketchup::Image && e.get_attribute("Animation", "folder", nil) && File.exist?(e.get_attribute("Animation", "folder", ""))
                images << e
            end#if
            if e.class==Sketchup::Group && e.get_attribute("Animation", "facemeID", nil)
                img=nil
                e.entities.each{|ee|
                    img=ee if ee.class==Sketchup::Image
                }
                images << img if img
            end#if
        }
        if ! images[0]
            UI.messagebox("ImageAnimator: Make_Unique Selected:\n\nYou must have one or more suitable Images Selected.\n [Images' Image-files Folders must have been preset]")
            return nil
        end#if
        @model.start_operation("Make_Unique")
        images.each{|image| image.make_unique() }
        @model.commit_operation
    end
    
    def clone()
        images=[]
        @ss.each{|e|
            if e.class==Sketchup::Image && e.get_attribute("Animation", "folder", nil) && File.exist?(e.get_attribute("Animation", "folder", ""))
                images << e
            end#if
        }
        gps=[]
        @ss.each{|e|
            if e.class==Sketchup::Group && e.get_attribute("Animation", "facemeID", nil)
                gps << e
            end#if
        }
        gimages=[]
        gps.each{|gp|
            cp=gp.copy ### duplicate faceme group
            img=nil
            cp.make_unique if cp.entities.parent.instances[1]
            cp.entities.each{|ee| img=ee if ee.class==Sketchup::Image }
            if img
              gimages << img
              id=Time.now.to_f+rand ### truly unique id
              img.set_attribute("Animation", "facemeID", id)
              cp.set_attribute("Animation", "facemeID", id)
            else
              cp.erase!
            end#if
        }
        if ! images[0] && ! gimages[0]
            UI.messagebox("ImageAnimator: Clone Selected:\n\nYou must have one or more suitable Images Selected.\n [Images' Image-files Folders must have been preset]")
            return nil
        end#if
        @model.start_operation("Clone")
        images.each{|image| image.clone() }
        gimages.each{|image| image.make_unique() }
        @model.commit_operation
    end
    
    
end#class

###########################################################################
### menu + toolbar
unless file_loaded?(__FILE__)
    ###
    icons=File.join(File.dirname(__FILE__), "ImageAnimatorIcons")
    ###
    toolbar=UI::Toolbar.new("ImageAnimator")
    ###
    cmd1=UI::Command.new("Set Folder for Selected Image"){ImageAnimator.new(1)}
	cmd1.small_icon=File.join(icons, "folder_small.png")
	cmd1.large_icon=File.join(icons, "folder_large.png")
	cmd1.tooltip="Set Folder for Selected Image"
    cmd1.status_bar_text="ImageAnimator: Set the Folder of Image-files for the Selected Image..."
	toolbar.add_item(cmd1)
    cmd8=UI::Command.new("Set Folder from Selected Image to 'Nil'"){ImageAnimator.new(8)}
	cmd8.small_icon=File.join(icons, "nofolder_small.png")
	cmd8.large_icon=File.join(icons, "nofolder_large.png")
	cmd8.tooltip="Set Folder from Selected Image to 'Nil'"
    cmd8.status_bar_text="ImageAnimator: Set the Folders fof Image-Files for the Selected Images to 'Nil'..."
	toolbar.add_item(cmd8)
    toolbar.add_separator
    ###
    cmd6=UI::Command.new("Set Timer for Selected Images"){ImageAnimator.new(6)}
	cmd6.small_icon=File.join(icons, "timer_small.png")
	cmd6.large_icon=File.join(icons, "timer_large.png")
	cmd6.tooltip="Set Timer for Selected Images"
    cmd6.status_bar_text="ImageAnimator: Sets the Timer for Selected Images..."
	toolbar.add_item(cmd6)
    ###
    cmd7=UI::Command.new("Set FaceMe for Selected Images"){ImageAnimator.new(7)}
	cmd7.small_icon=File.join(icons, "faceme_small.png")
	cmd7.large_icon=File.join(icons, "faceme_large.png")
	cmd7.tooltip="Set FaceMe for Selected Images"
    cmd7.status_bar_text="ImageAnimator: Sets the FaceMe Behavior for Selected Images..."
	toolbar.add_item(cmd7)
    toolbar.add_separator
    ###
    cmd2=UI::Command.new("Animate Selected Images"){ImageAnimator.new(2)}
	cmd2.small_icon=File.join(icons, "start_small.png")
	cmd2.large_icon=File.join(icons, "start_large.png")
	cmd2.tooltip="Animate Selected Images"
    cmd2.status_bar_text="ImageAnimator: Starts the Animation of the Selected Images..."
	toolbar.add_item(cmd2)
    ###
    cmd4=UI::Command.new("Stop Selected Images Animation"){ImageAnimator.new(4)}
	cmd4.small_icon=File.join(icons, "stop_small.png")
	cmd4.large_icon=File.join(icons, "stop_large.png")
	cmd4.tooltip="Stop Selected Images' Animation"
    cmd4.status_bar_text="ImageAnimator: Stops the Animation of the Selected Images..."
	toolbar.add_item(cmd4)
    toolbar.add_separator
    ###
    cmd3=UI::Command.new("Animate All Images"){ImageAnimator.new(3)}
	cmd3.small_icon=File.join(icons, "startall_small.png")
	cmd3.large_icon=File.join(icons, "startall_large.png")
	cmd3.tooltip="Animate All Images"
    cmd3.status_bar_text="ImageAnimator: Starts the Animation of All Suitable Images..."
	toolbar.add_item(cmd3)
    ###
    cmd5=UI::Command.new("Stop All Image Animation"){ImageAnimator.new(5)}
	cmd5.small_icon=File.join(icons, "stopall_small.png")
	cmd5.large_icon=File.join(icons, "stopall_large.png")
	cmd5.tooltip="Stop All Image Animation"
    cmd5.status_bar_text="ImageAnimator: Stops All Image Animation..."
	toolbar.add_item(cmd5)
    ###
    ### menu only tools...
    cmd9=UI::Command.new("Make_Unique Selected"){ImageAnimator.new(9)}
	cmd9.tooltip="Make_Unique Selected"
    cmd9.status_bar_text="ImageAnimator: The Selected Image(s) will be 'made unique'..."
	#toolbar.add_item(cmd9)
    ### menu only tools
    cmd10=UI::Command.new("Clone Selected"){ImageAnimator.new(10)}
	cmd10.tooltip="Clone Selected"
    cmd10.status_bar_text="ImageAnimator: The Selected Image(s) will be 'Cloned'..."
	#toolbar.add_item(cmd10)
    ###
    toolbar.restore if toolbar.get_last_state==TB_VISIBLE
    ###
    menu=UI.menu("View").add_submenu("ImageAnimator")
    menu.add_item(cmd1)
    menu.add_item(cmd8)
    menu.add_separator
    menu.add_item(cmd6)
    menu.add_item(cmd7)
    menu.add_separator
    menu.add_item(cmd2)
    menu.add_item(cmd4)
    menu.add_separator
    menu.add_item(cmd3)
    menu.add_item(cmd5)
    menu.add_separator
    menu.add_item(cmd9)
    menu.add_item(cmd10)
    ###
end#if
file_loaded(__FILE__)
###########################################################################
