=begin
###########################################################################
TIG (c) 2010 - 2013
  All Rights Reserved.
  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
  WARRANTIES,INCLUDING,WITHOUT LIMITATION,THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
###########################################################################
  ImageTrimmer.rb
###########################################################################
Needs 'image_class.rb' and 'EdgeDetector.jar' [in the SKM_tools folder]
###########################################################################
Adds a new Tool 'ImageTrimmer':

Menu > Tools > ImageTrimmer

It takes a Selected Image [it must be a .png file] with transparent areas 
and it finds all of those areas' boundaries and it adds 'Edges' to them.
It makes a new Component named after the Image containing faces with the 
Image applied as the Texture of a new Material with a matching name.
All transparent Faces are removed as 'trimmed' outlines or 'holes'.
The minimum size for a self-contained transparent area is 2x2 pixels - 
single transparent pixels or 'linear' pixels that are only a single pixel 
wide are merged into the solid areas.
If the Image contains no suitable transparent pixels the Tool exits early 
with a warning, with no changes made.
The processes show a % progress in the Command Line, it might take some 
time as there can be a lot of computation to do...
The number of edges in a 'Trimmed' Component is effectively the number of 
pixels on the transparent areas' boundaries - this can run into several 
thousand short edges!
An Instance of the new Component is placed at the Origin - it is initially 
made 'flat': you are asked if you want the new Component to 'Stand-up' - 
answer 'Yes' to rotate the Component's entities so it is vertical, or 'No' 
to leave it 'flat'.
You are asked if you want to 'Flatten' the new Component's Material Texture 
PNG [remove transparency] so it will receive shadows - 'Yes' to convert...
You are asked if you want to delete the original Image - answer 'Yes' to 
delete it, or 'No' to keep it.
A sub-folder is made in the Model's folder called 'TrimmedImages'.
The Tool uses it for the 'temporary' files needed to make its analysis and 
calculations.  On completion you are prompted to delete these temporary 
files - you answer 'Yes' to delete them, or 'No' to keep them - these files 
are named after the original Image - e.g. 'StandingMan1.png' = the original 
'imagefile' as extracted from the Image, 'StandingMan1[Edges].png' = a copy 
of the 'imagefile' but processed so that there is a one pixel 'black' 
outline of the transparent areas and all other areas colored 'white'.  
If you chose to 'flatten' the png there will also be another file 
'StandingMan1[Flattened].png' = a copy of the 'imagefile' but processed so 
there are no transparent areas.  You may find keeping these images useful ?
There is a temporary 'data' file 'StandingMan1[Edges].dat' which contains 
'01' data used to construct the Sketchup Edges - this should be auto-erased 
whatever option you choose.  If the 'TrimmedImages' folder is empty after 
the Tool completes it is deleted from the Model's folder.
You are asked if you want to 'Simplify the Outline', 'Cancel' to leave the 
outline with the maximum 'pixelation', or enter a 'Factor' and press OK to 
simplify the outline.
A very small 'Factor' will make few changes to the outlines [if any].
A 'Factor' ~1 will leave the outlines almost unchanged. 
A 'Factor' ~2 [default] will usually simplify an outline quite acceptably.
A 'Factor' ~3 will sometimes simplify the outline too much.
A higher 'Factor' [5-10] will often make a very simplified outline that 
might only be acceptable in exceptional circumstances.
This 'Simplification' is undoable.  Remember that you can always Select a 
Component later and Simplify its outline - see below...
The simplification processes show a % progress in the Command Line - the 
higher the 'Factor' you entered the quicker the outlines are simplified.

It is one/two/three/four/five step undo-able depending on what you chose 
to do:
undo the standing-up of the component, if changed.
undo the flattening of the png, if flattened.
undo the deletion of the original Image, if deleted.
undo simplification, if simplified.
undo the 'ImageTrim' completely.

The Tool is useful for auto-trimming transparent areas out of Images for 
'cardboard' cutouts of people etc.  After the Component is made you might 
want to change its origin/axes to say its bottom center and make it into a 
'face-me' component etc so it always faces the camera etc...
If you already have a Component using a transparent Material then save its 
Texture as an external Image file then Import as an Image, place over the 
original Component to resize appropriately and the run 'ImageTrimmer' on 
the Selected Image...


Menu > Tools > ImageTrimmer-Simplifier

After you have made 'Trimmed' Component you can simplify its outline by 
Selecting it and activating this tool.
If you do not have an acceptable Selection a dialog warns, the tool exits.
A dialog asks for the 'Factor, 'Cancel' to leave the outline unchanged or 
enter a 'Factor' and press OK to simplify the outline.
A very small 'Factor' will make few changes to the outlines [if any].
A 'Factor' ~1 will leave the outlines almost unchanged. 
A 'Factor' ~2 [default] will usually simplify an outline quite acceptably.
A 'Factor' ~3 will sometimes simplify the outline too much.
A higher 'Factor' [5-10] will often make a very simplified outline that 
might only be acceptable in exceptional circumstances.
This 'Simplification' is undoable.
The simplification processes show a % progress in the Command Line - the 
higher the 'Factor' you entered the quicker the outlines are simplified.
Idea: if 'simplifying' with a Factor of 2 produces a result that's not 
quite what you want, don't 'undo it' be re-simplify the simplified version 
with a Factor of 2 again, you might get a surprisingly improved outline...

###########################################################################
Donations:
  By Paypal.com to info @ revitrev.org
###########################################################################
Version:
1.0 20101216 First issue.
1.1 20101217 Simplification option and separate tool added.
1.2 20101220 MAC glitch fixed.
1.3 20101221 Time-out on file creation extended to 60secs or larger files.
1.4 20101222 MAC glitch on file paths addressed.
1.5 20101222 Time-out for larger dats increased, EdgeDetector.jar improved.
1.6 20110105 Time-out refined again.
1.7 20110105 Ignores very thin/small pixel areas and speedier processing.
1.8 20110111 Glitches with simpfify etc addressed.  Speed improved.
1.9 20110120 Option to 'flatten' material texture PNG so it receives shadows.
2.0 20110123 Flatten Image creation now waited for.
2.1 20110125 Simplified image - rare error on delete edge trapped.
2.2 20111013 MAC jar execution addressed.
2.3 20120827 Temp folder resolved.
###########################################################################
=end
###
require('sketchup.rb')
require(File.join(File.dirname(__FILE__), 'image_class.rb'))
###

class ImageTrimmer

	def initialize()
        @msg="ImageTrimmer: "
        Sketchup.status_text=@msg
        @model=Sketchup.active_model
        @path=@model.path
        if ! @path || @path==""
            UI.messagebox("ImageTrimmer:\n\nYou must Save the Model first.\n")
            return nil
        end#if
				###
				@path = SKMtools.encoder(@path)
				###
        @path=File.dirname(@path)
        @path.tr!("/","\\") if @pc
        @ss=@model.selection
        if ! @ss[0]
            UI.messagebox("ImageTrimmer:\n\nYou Selected an Image first.\n")
            return nil
        end#if
        @image=nil
        @ss.each{|e|
            if e.class==Sketchup::Image
                @image=e
                break
            end#if
        }
        if ! @image
            UI.messagebox("ImageTrimmer:\n\nYou Selected an Image first.\n")
            return nil
        end#if
        @pc=false
        @pc=true if RUBY_PLATFORM =~ /mswin|mingw/
        @defs=@model.definitions
        self.process_image()
    end

    def process_image()
        @name=@image.imagefile()
        if ! @name
            UI.messagebox("ImageTrimmer:\n\nCould not get Image's 'name'.\n")
            return nil
        end#if
        if ! File.extname(@name).downcase==".png"
            UI.messagebox("ImageTrimmer:\n\nImage must be a '.png' format file\nwith transparent areas.\n")
            return nil
        end#if
        @folder=File.join(@path, "TrimmedImages")
        if ! File.exist?(@folder)
            Dir.mkdir(@folder)
        elsif ! File.ftype(@folder)=="directory"
            UI.messagebox("ImageTrimmer:\n\nThis tool needs a folder\n  "+@folder+"\nBUT this already exists as a 'file' !\n")
            return nil
        end#if
        @folder.tr!("/","\\") if @pc
        @timage=@image.save_imagefile(@folder)
        @timage.tr!("/","\\") if @pc
		###
		@fol=SKMtools::FOLDER
		@temp=SKMtools::TEMP
		begin
			Dir.mkdir(@temp)
		rescue
		end
		###
        self.make_edges()
    end

    def make_edges()
        ###
        msg=@msg+"Making 'Edged' files..."
        Sketchup.status_text=msg
        ###
        base=File.join(File.dirname(@timage), File.basename(@timage, ".*"))
				base.tr!("/","\\") if @pc
        @eimage=base+"[Edges].png"
        @eimage.tr!("/","\\") if @pc
        @dimage=base+"[Edges].dat"
        @dimage.tr!("/","\\") if @pc
        File.delete(@eimage) if File.exist?(@eimage)
        File.delete(@dimage) if File.exist?(@dimage)
			### JAR hack for accents
			#p timage = @timage
			#p obase = base
			#@timage = timage.inspect.tr('"','').tr("\\\\","\\") if @pc
			#base = obase.inspect.tr('"','').tr("\\\\","\\") if @pc
			#p @timage
			#p base
		###
		jar=File.join(@fol, "EdgeDetector.jar").tr("\\","/")
		jartemp=File.join(@temp, "EdgeDetector.jar").tr("\\","/")
		  ###
		  file=File.open(jar, "rb")
			file.binmode
			data=file.read
		  file.close
		  ###
		  file=File.open(jartemp, "wb")
			file.binmode
			file.write(data)
		  file.close
		  ###
		  jar=jartemp
		###
        fromini=File.join(@temp, "from.ini").tr("\\","/")
        toini=File.join(@temp, "to.ini").tr("\\","/")
				fromini.tr!("/","\\") if @pc
        toini.tr!("/","\\") if @pc
        file=File.new(fromini, "w")
          file.write(@timage)
        file.close
        file=File.new(toini, "w")
          file.write(base)
        file.close
        ###
        File.chmod(0777, jar) if ! @pc
        if jar[0].chr=="/"
          jar="file:#{jar}"
        else
          jar="file:/#{jar}"
        end#if
        pwd=Dir.pwd
        Dir.chdir(@temp)
          UI.openURL(jar)
        Dir.chdir(pwd)
        ###
        msg=@msg+"Making 'Edges' files - please wait."
        Sketchup.status_text=msg
        ###
        wait=0.0
				#p @dimage
        until File.exist?(@dimage) || wait>=60.0 #!!!
            sleep(0.1)### wait till edged image dat is made
            wait+=0.1
            msg=msg+" ."
            Sketchup.status_text=msg
        end
        sleep(0.2)
        ### wait till all lines in dat file
        pheight=@image.pixelheight
        pwidth=@image.pixelwidth
        dheight=IO.readlines(@dimage).length
        dwidth=IO.readlines(@dimage)[0].length
        until dheight >= pheight && dwidth >= pwidth
          msg=msg+"  ."
          Sketchup.status_text=msg
          dheight=IO.readlines(@dimage).length
          dwidth=IO.readlines(@dimage)[0].length
        end#until
        sleep(1.0) ### to allow bigger dat files time to finish
		### remove temp .ini files
        File.delete(fromini) if File.exist?(fromini)
        File.delete(toini) if File.exist?(toini)
        ###
        if ! File.exist?(@dimage)
            UI.messagebox("ImageTrimmer:\n\nThe 'Edges' Image & Data files were NOT made !\n")
            return nil
        end#if
        self.add_edges()
        ###
    end

    def add_edges()
        begin
            @model.start_operation("ImageTrimmer", true)
        rescue
            @model.start_operation("ImageTrimmer")
        end
        pents=@image.parent.entities
        gp=pents.add_group
        ents=gp.entities
        msg=@msg+"Scanning Pixels: "
        Sketchup.status_text=msg
        dat=@dimage
        nlines=IO.readlines(dat)
        lines=[]
        nlines.each{|line|lines << line.chomp!}
        lines.compact!
        lines.reverse!
        height=lines.length
        width=lines[0].length
        tot=0
        (height).times{|y|
            tot+=1
            perc=100*tot/height
            Sketchup.status_text=msg+perc.to_s+"%"
            line=lines[y]
            next if ! lines[y][0]
            (width).times{|x|
                next if line[x]==48 || line[x]=='0'
                begin
                    ents.add_line([x,y,0],[x+1,y,0]) if lines[y][x+1]==line[x]#right
                rescue
                end
                begin
                    ents.add_line([x,y,0],[x,y+1,0]) if lines[y+1][x]==line[x]#up
                rescue
                end
                begin
                    ents.add_line([x,y,0],[x,y-1,0]) if lines[y-1][x]==line[x]#dn
                rescue
                end
            }
        }
        ### tidy up edges
        ents.to_a.each{|edge|
            next if ! edge.valid?
            edge.erase! if ! edge.start.edges[1] || ! edge.end.edges[1]
        }
        ents.to_a.each{|edge|
            next if ! edge.valid?
            edge.erase! if edge.start.edges[2] && edge.end.edges[2]
        }
        ents.to_a.each{|edge|
            next if ! edge.valid?
            edge.erase! if ! edge.start.edges[1] || ! edge.end.edges[1]
        }
        ### add corner markers
        ents.add_cpoint([0,0,0])
        ents.add_cpoint([width,0,0])
        ents.add_cpoint([width,height,0])
        ents.add_cpoint([0,height,0])
        ### componentize and transform to original etc
        compi=gp.to_component
        compd=compi.definition
        name=File.basename(@timage, ".*")
        compd.name=name
        name=compd.name
        ### match image
        cx=compi.bounds.width
        cy=compi.bounds.height
        ix=@image.width
        iy=@image.height
        x=ix/cx
        y=iy/cy
        trs=Geom::Transformation.scaling(ORIGIN, x, y, 0)
        ents=compd.entities
        ents.transform_entities(trs, ents.to_a)
        ###
        ents.to_a.each{|e|e.erase! if e.class==Sketchup::ConstructionPoint}
        ### face the edges and remove 'island holes'
        edges=ents.to_a ### all edges
        tot=edges.length
        while edges[0]
            perc=100*(tot-edges.length)/tot
            msg=@msg+"Processing Edges "+perc.to_s+"%"
            Sketchup.status_text=msg
            edges[0].find_faces
            edges.shift
            ents.to_a.each{|e|edges.delete(e) if e.class==Sketchup::Edge && e.faces[0]}
        end#while
        unfaced=true
        ents.each{|e|
            if e.class==Sketchup::Face
                unfaced=false
                break
            end#if
        }
        if unfaced
            Sketchup.status_text=""
            UI.messagebox("ImageTrimmer:\n\nThere were no Transparent Pixels\nthat could form Faces !\n")
            ents.erase_entities(ents.to_a)
            self.tidy()
            @model.commit_operation
            return nil
        end#if
        msg=@msg+"Removing Transparent Faces."
        Sketchup.status_text=msg
        ###
        faces=[]
        ents.to_a.each{|e|
          if e.class==Sketchup::Face
            e.reverse! if e.normal.z < 0
            faces << e
          end#if
        }
        ###
keptfaces=[]
flength=faces.length
(flength-1).times{|this|
  faces.each{|face|
    msg=msg+"."
    Sketchup.status_text=msg
    if face.valid?
      face.outer_loop.edgeuses.each{|edgeuse|
        if ! edgeuse.partners[0] ### outermost face
            faces=faces - [face]
            keptfaces << face
            loops=face.loops
            loops.each{|loop|
               msg=msg+"."
               Sketchup.status_text=msg
               faces.each{|fac|
                  msg=msg+"."
                  Sketchup.status_text=msg
                  if fac.valid? && (fac.outer_loop.edges - loop.edges)==[]
                    faces=faces - [fac]
                    fac.erase!### fac abutts kept face so it must be erased...
                    msg=msg+" "
                    Sketchup.status_text=msg
                  end#if fac
               }# fac
            }# loop
        end #if outermost
        break if keptfaces.length>=flength
      }# edgeuse
    else
      faces=faces - [face]
    end #if valid
  } #face
}#times
        faced=false
        ents.to_a.each{|e|
          if e.class==Sketchup::Face
            faced=true
            break
          end#if
        }
        if ! faced
          UI.messagebox("ImageTrimmer:\n\nIt is not possible to form Faces from this Image !\n")
          return nil
        end#if
        ### add image.material to comp'd faces
        msg=@msg+"Adding the Image to Faces - please wait..."
        Sketchup.status_text=msg
        ### make material from image
        mat=@model.materials.add(File.basename(@name, ".*"))
        mat.texture=@timage
        mat.texture.size=[@image.width, @image.height]
        ents.each{|e|
            if e.class==Sketchup::Face
                e.material=mat
                e.back_material=mat
            end#if
            if e.class==Sketchup::Edge
                e.hidden=true
            end#if
        }
        ###--------------------------------------------------
        @model.commit_operation
        ###
        Sketchup.status_text=""
        @ss.clear
        @ss.add(compi)
        ### make ahead of ourselves !!!!!!!!!!!
        ### NOW change from PNG to FLAT PNG
        fromini=File.join(@temp, "from.ini").tr("\\","/")
        toini=File.join(@temp, "to.ini").tr("\\","/")
        file=File.new(fromini, "w")
            file.write(@timage)
        file.close
        @flatPNG=@timage.gsub(/\.[Pp][Nn][Gg]$/,'[Flattened].png')
        file=File.new(toini, "w")
            file.write(@flatPNG)
        file.close
        jar=File.join(@fol, "FlattenPNG.jar").tr("\\","/")
		jartemp=File.join(@temp, "FlattenPNG.jar").tr("\\","/")
		  ###
		  file=File.open(jar, "rb")
			file.binmode
			data=file.read
		  file.close
		  ###
		  file=File.open(jartemp, "wb")
			file.binmode
			file.write(data)
		  file.close
		  ###
		  jar=jartemp
		###		
        File.chmod(0777, jar) unless @pc
        if jar[0].chr=="/"
          jar="file:#{jar}"
        else
          jar="file:/#{jar}"
        end#if
        pwd=Dir.pwd
        Dir.chdir(@temp)
          UI.openURL(jar)
        Dir.chdir(pwd)
        ###
        msg=@msg+"Flattening the 'PNG' - please wait."
        Sketchup.status_text=msg
        ###
        wait=0.0
        until File.exist?(@flatPNG) || wait>=60.0 #!!!
            sleep(0.1)### wait till flattened png is made
            wait+=0.1
            msg=msg+" ."
            Sketchup.status_text=msg
        end
        sleep(0.2)
        ###
        ### stand up?
		msg=@msg+"Make Component Stand-up ?"
        Sketchup.status_text=msg
        if UI.messagebox("Make Component Stand-up ?",MB_YESNO,"")==6 ### 6=YES 7=NO
            begin
                @model.start_operation("Make Component Stand-up", true)
            rescue
                @model.start_operation("Make Component Stand-up")
            end
            trr=Geom::Transformation.rotation(ORIGIN, X_AXIS, 90.degrees)
            ents.transform_entities(trr, ents.to_a)
            compd.bounds.clear ### sort bounds
            compd.bounds.add([0,0,0])
            compd.bounds.add([0,0,@image.height])
            begin
              compd.refresh_thumbnail
            rescue
            end
            @model.commit_operation
        end#if
        ###
        ### Ask PNG2JPG ?
        msg=@msg+"'Flatten' the PNG Texture of Component's Material [i.e. No Transparency] so that it will Receive Shadows ?"
        Sketchup.status_text=msg
		if UI.messagebox("'Flatten' the PNG Texture of Component's Material\n[i.e. No Transparency] so that it will Receive Shadows ?",MB_YESNO,"")==6 ### 6=YES 7=NO
        begin
            @model.start_operation("'Flatten' the PNG Texture", true)
        rescue
            @model.start_operation("'Flatten' the PNG Texture")
        end
        ###
        if File.exist?(@flatPNG)
          while mat.texture.filename != @flatPNG
           begin
            mat.texture=@flatPNG ###################
            mat.texture.size=[@image.width, @image.height]
           rescue
            mat.texture=@flatPNG ###################
            mat.texture.size=[@image.width, @image.height]
           end
          end#while
        else
          msg=@msg+"Flattening the 'PNG' - FAILED !"
          Sketchup.status_text=msg
        end#if
        ###
        @model.commit_operation
        File.delete(fromini) if File.exist?(fromini)
        File.delete(toini) if File.exist?(toini)
      else ### we don't want the flat file ?
        File.delete(@flatPNG) if File.exist?(@flatPNG)
      end#if
        ###
		msg=@msg+"Delete Original Image ?"
        Sketchup.status_text=msg
        ### delete original image?
        if UI.messagebox("Delete Original Image ?",MB_YESNO,"")==6 ### 6=YES 7=NO
            begin
                @model.start_operation("Delete Original Image", true)
            rescue
                @model.start_operation("Delete Original Image")
            end
            @image.erase! if @image.valid?
            @model.commit_operation
        end#if
        ### tidy
        self.tidy?()
        ###
        ### simplify
        ImageTrimmer.simplify(ents)
        ###
    end
    
    def tidy?()
        ### delete temp files?
		msg=@msg+"Delete Temporary 'ImageTrimmer' Files ?"
        Sketchup.status_text=msg
        File.delete(@dimage) if File.exist?(@dimage)### always remove dat
        if UI.messagebox("Delete Temporary 'ImageTrimmer' Files ?",MB_YESNO,"")==6 ### 6=YES 7=NO
            self.tidy()
        end#if
    end
    
    def tidy()
        File.delete(@dimage) if File.exist?(@dimage)
        File.delete(@timage) if File.exist?(@timage)
        File.delete(@eimage) if File.exist?(@eimage)
        File.delete(@flatPNG) if File.exist?(@flatPNG)
        Dir.delete(@folder) if Dir.entries(@folder).length <= 2
    end

    ##### simplify after made
    def ImageTrimmer.simplifier()
        model=Sketchup.active_model
        sel=model.selection
        if ! sel[0] || sel[1]
            UI.messagebox("Simplify Outlines:\n\nMust Select ONE Component-Instance.\n")
            return nil
        end#if
        ci=sel[0]
        if ci.class==Sketchup::ComponentInstance
            ents=ci.definition.entities
        else
            UI.messagebox("Simplify Outlines:\n\nMust Select ONE Component-Instance.\n")
            return nil
        end#if
        return nil if ! ents[0]
        ###
        ImageTrimmer.simplify(ents)
        ###
    end
    
  ########### based on ideas from Thomthom and others..........
  def ImageTrimmer.simplify(ents)
    ###
    @msg="ImageTrimmer-Simplifier: "
    msg=@msg+"Factor ?"
    Sketchup.status_text=msg
    model=Sketchup.active_model
    faces=[]
    ents.each{|e|faces << e if e.class==Sketchup::Face}
    loops=[]
    fmat=nil
    bmat=nil
    faces.each{|face|
      face.loops.each{|loop|
        loops << loop if ! loops.include?(loop)
      }
      fmat=face.material if face.material
      bmat=face.back_material if face.back_material
    }
    vsets=[]; loops.each{|loop|
      vsets << [loop.edges, loop.vertices]
    }
    # UI = Epsilon - smaller = more detail, 1=as_was, 2 best, 7 !
    prompts = ["Factor: "]
    defaults = [2.0]
    results = UI.inputbox(prompts, defaults, 'ImageTrimmer-Simplifier')
    Sketchup.status_text="" if ! results
    return nil if ! results
    ###
    epsilon = results[0].to_f
    epsilon = 0.1 if epsilon <= 0.0
    epsilon = epsilon / 10.0
	epsilon = epsilon * 2 if Sketchup.version.to_i > 13
    begin
      model.start_operation('ImageTrimmer-Simplifier', true)
    rescue
      model.start_operation('ImageTrimmer-Simplifier')
    end
    msg=@msg+"Please wait..."
    Sketchup.status_text=msg
    ents.erase_entities(faces)
    tgp=ents.add_group()
    tents=tgp.entities
    alledges=[]
    tot=vsets.length
    ctr=1
    vsets.each{|vset|
      perc=100*ctr/tot
      ctr+=1
      msg=@msg+"Simplifying Loops #{perc}%"
      Sketchup.status_text=msg
      vertices=vset[1]
      points = vertices.collect{|v|v.position}
      simplified_curve = ImageTrimmer.douglas_peucker(points, epsilon)
      simplified_curve << simplified_curve[0] ### so it loops
      edges=tents.add_curve(simplified_curve)
      edges.each{|e|e.hidden=true}
      alledges << edges
    }
    alledges.flatten!
    ### remove old edges
    vsets.each{|vset| vset[0].each{|e| e.erase! if e.valid?} }
    tgp.explode
    ### make faces etc...
    edges=alledges ### all edges
        tot=edges.length
        while edges[0]
            perc=100*(tot-edges.length)/tot
            msg=@msg+"Processing Edges #{perc}%"
            Sketchup.status_text=msg
            edges[0].find_faces if edges[0].valid?
            edges.shift
            ents.to_a.each{|e|edges.delete(e) if e.class==Sketchup::Edge && e.faces[0]}
        end#while
        unfaced=true
        ents.each{|e|
            if e.class==Sketchup::Face
                unfaced=false
                break
            end#if
        }
        if unfaced
            Sketchup.status_text=""
            UI.messagebox("ImageTrimmer-Simplify:\n\nThere were no Shapes\nthat could form Faces !\n")
            model.commit_operation
            return nil
        end#if
        msg=@msg+"Removing 'Holes'."
        Sketchup.status_text=msg
        ###
        faces=[]
        ents.to_a.each{|e|
          if e.class==Sketchup::Face
            e.reverse! if e.normal.z < 0
            e.reverse! if e.normal.y > 0
            faces << e
          end#if
        }
        ###
keptfaces=[]
flength=faces.length
(flength-1).times do |this|
  faces.each{|face|
    msg=msg+"."
    Sketchup.status_text=msg
    if face.valid?
      face.outer_loop.edgeuses.each{|edgeuse|
        if ! edgeuse.partners[0] ### outermost face
            faces=faces - [face]
            keptfaces << face
            loops=face.loops
            loops.each{|loop|
               msg=msg+"."
               Sketchup.status_text=msg
               faces.each{|fac|
                  msg=msg+"."
                  Sketchup.status_text=msg
                  if fac.valid? && (fac.outer_loop.edges - loop.edges)==[]
                    faces=faces - [fac]
                    fac.erase!### fac abutts kept face so it must be erased...
                    msg=msg+" "
                    Sketchup.status_text=msg
                  end#if fac
               }# fac
            }# loop
        end #if outermost
        break if keptfaces.length>=flength
      }# edgeuse
    else
      faces=faces - [face]
    end #if valid
  } #face
end#times       
        ### add image.material to comp'd faces
        msg=@msg+"Adding the Image to Faces - please wait..."
        Sketchup.status_text=msg
        ents.each{|e|
            if e.class==Sketchup::Face
                e.material=fmat
                e.back_material=bmat
            end#if
            if e.class==Sketchup::Edge
                e.hidden=true
            end#if
        }
        ###
        Sketchup.status_text=""
    ###
    model.commit_operation
    ###
  end
  
  ###----------------------------------------------------------------------
  ### http://en.wikipedia.org/wiki/Ramer-Douglas-Peucker_algorithm
  ### http://en.wiki.mcneel.com/default.aspx/McNeel/PolylineSimplification.html
  def ImageTrimmer.douglas_peucker(points, epsilon)
    ###
    return points if points.length < 3
    ### Find the point with the maximum distance
    dmax = 0
    index = 0
    line = [points.first, points.last]
    1.upto(points.length - 2) { |i|
      d = points[i].distance_to_line(line)
      if d > dmax
        index = i
        dmax = d
      end
    }
    ### If max distance is greater than epsilon, recursively simplify
    result = []
    if dmax >= epsilon
      ### Recursive call
      recResults1 = ImageTrimmer.douglas_peucker(points[0..index], epsilon)
      recResults2 = ImageTrimmer.douglas_peucker(points[index...points.length], epsilon)
      ### Build the result list
      result = recResults1[0...-1] + recResults2
    else
      result = [points.first, points.last]
    end#if
    ###
    return result
    ###
  end

end#class

###########################################################################
### menu
unless file_loaded?(__FILE__)
    UI.menu("Tools").add_item("ImageTrimmer"){ImageTrimmer.new()}
    UI.menu("Tools").add_item("ImageTrimmer-Simplifer"){ImageTrimmer.simplifier()}
end#if
file_loaded(__FILE__)
###########################################################################
