=begin
TIG (c) 2010 - 2013
  All Rights Reserved.
  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
  WARRANTIES,INCLUDING,WITHOUT LIMITATION,THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
[NB:definition and transformation methods are Copyright 2008, AlexM]
###########################################################################
  image_class.rb
###########################################################################
Needs 'Bumpmapper.jar, 'BumpmapperInverted.jar', 'Thumbnailer.jar' and 
'ThumbnailerRGB.jar' in the same sub-folder.
You will need Java installed on your PC/MAC - it usually is...
Also may also needs certain other .rb's that come in the same folder.
Temporary .ini files are written to the ../Plugins/SKMtools/ folder and
are deleted by the various methods to pass values to the .jar executable.
Loaded by 'SKMtools_loader.rb' in the ../Plugins/ folder.
###########################################################################
Adds new methods to the Sketchup class 'Image' - including:

Image.definition       
>>> returns the 'definition' of the image.

Image.transformation
>>> returns the 'transformation' of the image.

Image.transformation=tr
>>> set the 'transformation' of the image to a new transformation 'tr'.

Image.use_as_material
>>> makes a new material based on the Image AND keeps the Image intact, 
    it 'puts' the new Material name and it returns the new Material.
    
Image.make_unique
>>> if there is more that one instance of an Image the specified image is 
    'made unique' - now any subsequent changes to it will not affect others.
    The associated image_file is duplicated and its name is incremented 
    with a numerical suffix.
    Any Attributes [Animation:folder/timer etc] are cloned from the 
    original Image into the now 'unique' version.
    It returns a new Image object, so you will need to use 
    'image=image.make_unique' to retain the reference to it as 'image'...
    
Image.copy
>>> this makes a copy of the Image using the same definition.
    Any Attributes [Animation:folder/timer etc] are kept in the copy.
    It returns the new Image object for the copied Image so use 
    'copiedimage=image.copy' to get its reference for later...
    
Image.clone
>>> this is like Image.copy BUT it runs an Image.make_unique on it so 
    it is a separate definition of a new Image object.
    The associated image_file is duplicated and its name is incremented 
    with a numerical suffix.
    Any Attributes [Animation:folder/timer etc] are cloned from the 
    original Image into the now 'unique' version.
    It returns a new Image object, so you will use 
    'clonedimage=image.clone' to get the reference to it for later...
    
Image.imagefile
>>> returns the name of the Image's image-file - e.g. 'Poster666.png'
    
Image.imagefile='filepath'
>>> this changes the image-file used by the Image to the image 'filepath', 
    it returns the 'imagefile-name' if successful, or 'nil'.
    Only the following file-types are useable:
    ".jpg", ".jpeg", ".png", ".psd", ".tif", ".tga" and ".bmp".
    
Image.save_imagefile('folder') 
>>> saves the image-file used by the Image into 'folder', 
    it keeps the original image file's name [with spaces etc squeezed out], 
    returns the path the the image-file - e.g. "C:\\Temp\\Poster_01.png".

Image.folder='folder'
>>> this sets the 'folder' that the Image might use for an 'animation'.
    It defaults to 'nil'.  It is saved as an Attribute with the Image - 
    'Animation','folder'.  When used later, if the 'folder' is not found, 
    then it is reset to 'nil'.
    To 'unset' an Image's folder use Image.folder=nil
    The 'folder' should contain image-files that are numbered in the order 
    you wish them to be used e.g. img001.png, img002.png, img003.png etc.
    Only files with the following file-type suffixes are used:
      ".jpg", ".jpeg", ".png", ".psd", ".tif", ".tga" and ".bmp"
    To use the the 'folder' you invoke to use on of the 'animation' 
    methods explained below.
    
Image.folder
>>> returns the 'folder' associated with the Image.
    Returns 'nil' if not set: of if not found it's reset to 'nil'.
    
Image.timer='timer'
>>> by default any Image with a 'folder' set has its 'timer' set to 'nil'.
    You can reset it to any number >= 0.0 using this method, e.g. 
    Image.timer=0.1.  If the 'timer' is 'nil' [0.0] then the Image's 
    'animation' will swap its image-files over only when a Scene tab 
    changes - e.g. during a scene-animation sequence.  However, if it's set 
    to > 0.0 then the Image will also animate within the Scene as you 
    work, pan/zoom etc and also across scene-animations etc. 
    This is invoked by a UI.timer() changing frames every 0.1 seconds.
    The 'timer' is saved as an Attribute of the Image ["Animation","timer"].
    If the Image has no 'folder' set it's set to 'nil'.
    It returns the float of the 'timer' value as set, or 'nil'.

Image.timer
>>> return a 'float' representing the 'timer' set for that Image.
    It is read from the Image's Attribute "Animation","timer".
    If it's not set it returns 'nil'.
    If the Image has no 'folder' set it's set to 'nil'.
    
Image.faceme='type'
>>> this sets the 'faceme' Attribute of the Image to 'type'.  
    If 'nil' [or '0'] it stops the Image's 'faceme' property - it keeps the 
    Transformation that the Image has when the 'faceme' property ends.
    If '1' it keeps the Image's origin to its bottom-left and makes it 
    always face the camera, rotating about the Z_AXIS as it does so.
    If '2' it resets the Image's origin to its bottom-center and makes it 
    always face the camera, rotating about the Z_AXIS as it does so.
    If '3' it resets the Image's origin to its bottom-right and makes it 
    always face the camera, rotating about the Z_AXIS as it does so.
    Every Instance of the Image in placed into its own Group which has its 
    'faceme' behaviour set 'true'.
    If the Image's 'faceme' is set to 'nil' [or '0'] then any 'faceme' 
    Groups that contain the Image's Instances are exploded.
    The Image/Group 'faceme' property and links are remembered across sessions.
    
Image.faceme
>>> returns 'nil', 1, 2 or 3 depending on the Image's 'faceme' Attribute.
    
Image.start_animation
>>> It uses each of the image-files found in the Images's 'folder' - 
    as set by Image.folder='folder' as explained above.
    If the Image does not have an 'Animation','folder' Attribute set then 
    it returns 0.  Assuming the 'folder' exists then its image-files are 
    used in turn, in alphanumerical order [img01.png, img02.png, etc], and 
    the 'Image.imagefile=filepath' method changes the Image's 'image-file'.  
    All of the image-files in the 'set' found in the 'folder' should 
    preferably have the same pixel w/h etc to ensure consistency and the 
    same file-type [e.g. all .png] to avoid 'loading-order' discrerencies.
    If only one instance of the the Image is to be changed then remember to 
    use 'Image.make_unique' on it to separate it from all other instances 
    before using this tool.  The Image now displays each of the specified 
    image-files in turn as the Image's 'image-file'.
    
    If the Image's 'timer' is set to 'nil' then the Image's image-file 
    changes with every change of 'Scene' - when makng an Animation etc 
    [an Observer is added to the Model's Pages to facilitate this automatic 
    change]. The list of image-files is processed repeatedly until either 
    the methods Image.stop_animation or Image.stop_animation_all are 
    invoked.  If you add a new Scene during an animation sequence [and it 
    is set as 'included' in the animation in the settings in the 
    Scene-Manager] then these should be included in the Image's animation 
    sequence when it is next becomes the 'active' Scene.
    The rate of change of the Scene tabs during an Animation will clearly 
    affect the appearance of the Image animation changes...
    
    If the Image's 'timer' is set to a number [of seconds] then the Image's 
    image-file changes every 'timer' seconds - as you work or make scene-
    animations. The list of image-files is processed repeatedly until 
    either the methods Image.stop_animation or Image.stop_animation_all are 
    invoked.
    
    You can use this tool to animate Images as backgrounds-through-windows, 
    Computer/TV-screens, flickering-fires/flames, waving-flags, rotating-
    wheels,/machinery, flapping-birds, walking-people/animals 
    [when combined with multiple instances of the Image in different 
    locations and on different layers on/off between scenes to mimic the 
    'translation' movement as well as the repetitive movement within the 
    Image animation itself] etc.  
    If you are moving the Camera between each Scene, if desired you can set 
    the Image to 'faceme' if you want the Image to always face the viewer, 
    e.g. a flickering-candle-flame: otherwise it can be left, e.g. 
    a TV-screen-image placed onto the TV-set.
    Use the Image.faceme='type' method to [re]set this property, see above.
    
    Returns 'true' if set or 'false' if not set, e.g. the Image's 'folder' 
    is no longer available or it contains no valid images.
    If the Image's 'folder' is not found then the Image's Attribute 
    'Animation','run' is reset to 'false'.
    
    Image animation effects only work during that session and have to be 
    restarted when the Model is reopened. The Image's 'folder' Attribute is 
    remembered across sessions.
    
    Image.start_animation will stop any animations already running on 
    that Image - i.e. an Image can only have one active animation at a time.
    
    The 'set' of image-files used is a snap-shot of the contents of the 
    'folder' at the moment the animation is started, so any changes to the 
    'set' are not registered until a new Image.start_animation is 
    invoked, stopping the current one and then running the new version 
    using the image-files 'set' currently in the 'folder'.
    
    Returns 1 if the animation was successfully started, 0 otherwise.
    
Image.start_animation_all
>>> This finds all Images with an 'Animation','folder' Attribute set and 
    then starts their animation - i.e. a Pages Observer is amde for each 
    Image. See Image.start_animation above for more details...
    Returns the number of animations started.
    
Image.stop_animation
>>> this stops the current animation [if any].  
    It either removes the Observer from the Model's Pages so that the 
    Image does not display 'animation' effects or if the 'timer' is > 0.0 
    it cancels the current UI.timer() and the resultant 'animation'.
    The Image is 'frozen' using its current image-file.
    If you want to revert to a particular image-file then after 'stopping' 
    the animation sequence you will need to invoke the method to change it 
    using Image.imagefile='path_to_defaultimage.png'.
    Returns 1 if the animation was successfully stopped, 0 otherwise.
    
Image.stop_animation_all
>>> this stops ALL currently running Images' animation effects.
    It sets the ALL Images' Attribute 'Animation','run' to be 'false'.
    It either removes all of the Observers from the Model's Pages so that 
    NO Images will display 'animation' effects or it cancels the all of the 
    UI.timers and their resultant 'animations'.
    The Images are 'frozen' using their current image-file.
    If you want to revert to a particular image-file for a particular Image 
    then after 'stopping_all' you will need to invoke the method to change 
    it using Image.imagefile='path_to_defaultimage.png'.
    Returns the number of animations stopped.
    
Image.xflipped?
    Returns true if flipped in X.
    
Image.yflipped?
    Returns true if flipped in Y.
    
###########################################################################
There is a separate Tool 'ImageAnimator' - this uses these methods, using a 
toolbar interface to affect a selected Image and set/reset its 'folder'; it 
also 'starts' and 'stops' the animation of the selected Image or all-images.
It is packaged with this file - see notes within it [ImageAnimator.rb] 
###########################################################################
Donations:
  By Paypal.com to info @ revitrev.org
###########################################################################
Version:
1.0 20101209 First issue with extended toolset - animation etc.
1.1 20110106 Image.transformation= now takes a transformation OR array.
1.2 20110106 Robustness improved with existing method checks...
1.3 20110225 Minor glitch with ImageAnimationAppObserver fixed.
1.4 20130827 Temp folder fixed.
TODO Image.xflipped? and Image.yflipped?
             
###########################################################################
=end
###
require('sketchup.rb')
###

class Sketchup::Image

	unless Sketchup::Image.method_defined?(:definition)
      def definition()
        if not self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
		self.model.definitions.each { |d| return d if d.image? && d.instances.include?(self)}
        return nil
	  end#def
    end#if

    unless Sketchup::Image.method_defined?(:transformation)
	  def transformation()
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
		origin = self.origin
		axes = self.normal.axes
		tr = Geom::Transformation.axes(ORIGIN, axes.at(0), axes.at(1), axes.at(2))
		tr = tr*Geom::Transformation.rotation(ORIGIN, Z_AXIS, self.zrotation)
		tr = (tr*Geom::Transformation.scaling(ORIGIN, self.width/self.pixelwidth, self.height/self.pixelheight, 1)).to_a
		tr[12] = origin.x
		tr[13] = origin.y
		tr[14] = origin.z
		return Geom::Transformation.new(tr)
	  end#def
    end#if

	unless Sketchup::Image.method_defined?(:transformation=)
      def transformation=(tr)
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        if tr.class==Array
          tr=Geom::Transformation.new(tr)
        end#if
		status=self.transform!(self.transformation.inverse * tr)
        if status
          return self
        else
          return nil
        end#if
	  end
    end#if

    unless Sketchup::Image.method_defined?(:use_as_material)
      def use_as_material()
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        model=self.model
        mats=model.materials.to_a
        temp_group=model.active_entities.add_group()
        tr=self.transformation
        ins=temp_group.entities.add_instance(self.definition, tr)
        ins.explode
        temp_group.erase! if temp_group.valid?
        mats=model.materials.to_a - mats
        mat=mats[0]
        puts mat.display_name
        return mat
      end
    end#if
    
    unless Sketchup::Image.method_defined?(:make_unique)
      def make_unique()
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        defn=self.definition
        if defn.instances.length < 2
            puts "The Image is already unique!"
            return self
        end#if
        ###
        model=self.model
        idefs=[]
        model.definitions.each{|d|idefs << d if d.image?}
        inames=[]
        idefs.each{|idef|
            face=nil
            idef.entities.each{|e|
                if e.class==Sketchup::Face
                    face=e
                    break
                end#if
            }
            mat=face.material
            tex=mat.texture
            name=File.basename(tex.filename)
            inames << name
        }
		###
        temp=SKMtools::TEMP
		begin
			Dir.mkdir(temp)
		rescue
		end
        timg=self.save_imagefile(temp)
		###
        return self if ! timg
        ### make unique name
        tname=File.basename(timg)
        dirn=File.dirname(timg)
        extn=File.extname(timg)
        base=File.basename(tname, ".*")+"0"
        while inames.include?(tname)
          base.next!
          tname=base+extn
        end#if
        tpath=File.join(dirn, tname)
        if ! timg==tpath
          File.rename(timg, tpath)
          timg=tpath
        end#if
        ###
        w=self.pixelwidth
        h=self.pixelheight
        ents=self.parent.entities
        img=ents.add_image(timg, ORIGIN, w, h)
        File.delete(timg) if File.exist?(timg)
        return self if ! img
        tr=self.transformation
        img.transformation=tr
        ### clone attributes
        atds=self.attribute_dictionaries
        atds.each{|atd|
            atd.each_pair{|k,v|
                img.set_attribute(atd.name, k, v)
            } if atd
        } if atds
        ###
        self.erase!
        ###
        return img
      end
    end#if

    unless Sketchup::Image.method_defined?(:copy)
      def copy()
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        model=self.model
        ents=self.parent.entities
        gp=ents.add_group(self)
        cp=gp.copy
        gp.explode
        img=cp.entities[0]
        cp.explode
        return img
      end
    end#if
       
    unless Sketchup::Image.method_defined?(:clone)
      def clone()
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        model=self.model
        idefs=[]
        model.definitions.each{|d|idefs << d if d.image?}
        inames=[]
        idefs.each{|idef|
            face=nil
            idef.entities.each{|e|
                if e.class==Sketchup::Face
                    face=e
                    break
                end#if
            }
            mat=face.material
            tex=mat.texture
            name=File.basename(tex.filename)
            inames << name
        }
        temp=SKMtools::TEMP
		begin
			Dir.mkdir(temp)
		rescue
		end
        timg=self.save_imagefile(temp)
        return nil if ! timg
        ###
        w=self.pixelwidth
        h=self.pixelheight
        ents=self.parent.entities
        img=ents.add_image(timg, ORIGIN, w, h)
        File.delete(timg) if File.exist?(timg)
        return nil if ! img
        tr=self.transformation
        img.transformation=tr
        ### clone attributes
        atds=self.attribute_dictionaries
        atds.each{|atd|
            atd.each_pair{|k,v|
                img.set_attribute(atd.name, k, v)
            } if atd
        } if atds
        ###
        return img
        ###
      end
    end#if
     
    unless Sketchup::Image.method_defined?(:imagefile)
      def imagefile()
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        return nil if ! defn=self.definition
        face=nil
        defn.entities.each{|e|
          if e.class==Sketchup::Face
            face=e
            break
          end#if
        }
        return nil if ! face
        return nil if ! mat=face.material
        return nil if ! tex=mat.texture
        return nil if ! filename=File.basename(tex.filename)
        return filename
      end
    end#if
    
    unless Sketchup::Image.method_defined?(:imagefile=)
      def imagefile=(path="")
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        if ! path.class==String
            path=""
        end#if
				###
				path = SKMtools.encoder(path)
				###
        if path==""
            puts "Usage: image.set_imagefile('path_to_imagefile')"
            return nil
        end#if
        if not File.exist?(path)
            puts "Usage: image.set_imagefile('path_to_imagefile')"
            puts "File '#{path}' was not found."
            return nil
        end#if
        if File.ftype(path)=="directory"
            puts "Usage: image.set_imagefile('path_to_imagefile')"
            puts "Folder '#{path}' is a 'Directory'."
            puts "It needs an image format file like:"
            puts '[".jpg",".jpeg",".png",".psd",".tif",".tga",".bmp"]'
            return nil
        end#if
        imgs=[".jpg",".jpeg",".png",".psd",".tif",".tga",".bmp"]
        basename=File.basename(path)
        extname=File.extname(basename).downcase
        if ! imgs.include?(extname)
            puts "Usage: image.set_imagefile('path_to_imagefile')"
            puts "It needs an image format file like:"
            puts '[".jpg",".jpeg",".png",".psd",".tif",".tga",".bmp"]'
            return nil
        end#if
        ###
        return nil if ! defn=self.definition
        face=nil
        defn.entities.each{|e|
          if e.class==Sketchup::Face
            face=e
            break
          end#if
        }
        return nil if ! face
        ### get extg texture info
        tex=face.material.texture
        w=tex.width
        h=tex.height
        ### rest texture
        face.material.texture=path
        tex=face.material.texture
        ### resize to suit
        tex.size=[w, h]
        ###
        if tex
            return tex
        else
            return nil
        end#if
      end
    end#if
    
    unless Sketchup::Image.method_defined?(:save_imagefile)
      def save_imagefile(path="")
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        if ! path.class==String
            path=""
        end#if
				###
				path = SKMtools.encoder(path)
				###
        if path==""
            puts "Usage: image.save_imagefile('folder')"
            return nil
        end#if
        if ! File.exist?(path)
            puts "Usage: image.save_imagefile('folder')"
            puts "Folder '#{path}' was not found."
            return nil
        end#if
        if not File.ftype(path)=="directory"
            puts "Usage: image.save_imagefile('folder')"
            puts "Folder '#{path}' is not a 'Directory'."
            return nil
        end#if
        ###
        model=self.model
        ###
        pc=false
        pc=true if RUBY_PLATFORM =~ /mswin|mingw/
        ###
        return nil if ! defn=self.definition
        face=nil
        defn.entities.each{|e|
          if e.class==Sketchup::Face
            face=e
            break
          end#if
        }
        return nil if ! face
        return nil if ! mat=face.material
        tw=Sketchup.create_texture_writer
        ###
        temp_group=model.active_entities.add_group()
        temp_group.material=mat
        iname=File.basename(mat.texture.filename).gsub(/[^\.A-Za-z0-9_-]/, "_")
        ### in case there a folder path
        ipath=File.join(path, iname)
        tw.load(temp_group)
        tw.write(temp_group, ipath)
        temp_group.erase! if temp_group.valid?
        ipath.tr!("/","\\") if pc
        ###
        if File.exist?(ipath)
            return ipath
        else
            return nil
        end#if
      end
    end#if
    
    unless Sketchup::Image.method_defined?(:folder=)
      def folder=(path=nil)
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
				###
				path = SKMtools.encoder(path)
				###
        if path==""
            puts "Usage: image.folder='folder'"
            path=nil
        end#if
        if path && ! File.exist?(path)
            puts "Usage: image.folder='folder'"
            puts "Folder '#{path}' was not found."
            path=nil
        end#if
        if path && ! File.ftype(path)=="directory"
            puts "Usage: image.folder='folder'"
            puts "Folder '#{path}' is not a 'Directory'."
            path=nil
        end#if
        self.set_attribute("Animation", "folder", path)
        timer=self.get_attribute("Animation", "timer", nil)
        timer=nil if ! timer
        timer=nil if ! folder
        self.get_attribute("Animation", "timer", timer)
      end
    end#if
    
    unless Sketchup::Image.method_defined?(:folder)
      def folder()
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        folder=self.get_attribute("Animation", "folder", nil)
				###
				folder = SKMtools.encoder(folder)
				###
        folder=nil if folder && ! File.exist?(folder)
        timer=self.get_attribute("Animation", "timer", nil)
        timer=nil if ! folder
        self.set_attribute("Animation", "folder", folder)
        self.set_attribute("Animation", "timer", timer)
        return folder
      end
    end#if
    
    unless Sketchup::Image.method_defined?(:timer=)
      def timer=(timer=nil)
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        timer=timer.to_f if timer
        timer=nil if timer && timer <= 0.0
        #timer=0.1 if timer && timer < 0.1
        folder=self.get_attribute("Animation", "folder", nil)
				###
				folder = SKMtools.encoder(folder)
				###
        folder=nil if folder && ! File.exist?(folder)
        timer=nil if ! folder
        self.set_attribute("Animation", "timer", timer)
        return timer
      end
    end#if
    
    unless Sketchup::Image.method_defined?(:timer)
      def timer()
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        folder=self.get_attribute("Animation", "folder", nil)
				###
				folder = SKMtools.encoder(folder)
				###
        folder=nil if folder && ! File.exist?(folder)
        timer=self.get_attribute("Animation", "timer", nil)
        timer=nil if ! folder
        self.set_attribute("Animation", "folder", folder)
        self.set_attribute("Animation", "timer", timer)
        return timer
      end
    end#if
   
    unless Sketchup::Image.method_defined?(:faceme=)
      def faceme=(type=nil)
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        type=nil if ! [nil,1,2,3].include?(type)
        otype=self.faceme()
        defn=self.definition
        ints=defn.instances
        ###
        if type==1
            if otype==1
                ### do nothing
            elsif otype==2
                self.faceme=nil
                self.faceme=1
            elsif otype==3
                self.faceme=nil
                self.faceme=1
            else ### nil
                ints.each{|inst|
                    gp=inst.parent.entities.add_group(inst)
                    gp.definition.behavior.always_face_camera=true
                    inst.set_attribute("Animation", "faceme", 1)
                    id=Time.now.to_f+rand
                    inst.set_attribute("Animation", "facemeID", id)
                    gp.set_attribute("Animation", "facemeID", id)
                    gp.name="FaceMeImage"
                }
            end#if    
        elsif type==2
            if otype==2
                ### do nothing
            elsif otype==1
                self.faceme=nil
                self.faceme=2
            elsif otype==3
                self.faceme=nil
                self.faceme=2
            else ### nil
                ints.each{|inst|
                    gp=inst.parent.entities.add_group(inst)
                    ### reset gp origin to BM
                    width=gp.bounds.width
                    height=gp.bounds.height
                    vec=Geom::Vector3d.new(-width/2, -height/2, 0)
                    tr=Geom::Transformation.translation(vec)
                    inst.transform!(tr)
                    gp.transform!(tr.inverse)
                    ###
                    gp.definition.behavior.always_face_camera=true
                    inst.set_attribute("Animation", "faceme", 2)
                    id=Time.now.to_f+rand
                    inst.set_attribute("Animation", "facemeID", id)
                    gp.set_attribute("Animation", "facemeID", id)
                    gp.name="FaceMeImage"
                }
            end#if    
        elsif type==3
            if otype==3
                ### do nothing
            elsif otype==1
                self.faceme=nil
                self.faceme=3
            elsif otype==2
                self.faceme=nil
                self.faceme=3
            else ### nil
                ints.each{|inst|
                    gp=inst.parent.entities.add_group(inst)
                    ### reset gp origin to BR
                    width=gp.bounds.width
                    height=gp.bounds.height
                    vec=Geom::Vector3d.new(-width, -height, 0)
                    tr=Geom::Transformation.translation(vec)
                    inst.transform!(tr)
                    gp.transform!(tr.inverse)
                    ###
                    gp.definition.behavior.always_face_camera=true
                    inst.set_attribute("Animation", "faceme", 3)
                    id=Time.now.to_f+rand
                    inst.set_attribute("Animation", "facemeID", id)
                    gp.set_attribute("Animation", "facemeID", id)
                    gp.name="FaceMeImage"
                }
            end#if    
        else ### it's 'nil' so explode all groups
            if otype ### i.e. it's not already 'nil'
                model=self.model
                defs=model.definitions
                gps=[]
                defs.each{|defn| gps << defn.instances if defn.group?}
                gps.flatten!
                fgps=[]
                gps.each{|gp|fgps << gp if gp.get_attribute("Animation", "facemeID", nil)}
                ints.each{|inst|
                    inst.set_attribute("Animation", "faceme", nil)
                    id=inst.get_attribute("Animation", "facemeID", nil)
                    inst.set_attribute("Animation", "facemeID", nil)
                    next if ! id
                    fgps.each{|gp|gp.explode if gp.valid? and id==gp.get_attribute("Animation", "facemeID", nil)}
                }
            end#if
        end#if
        return type
      end
    end#if
    
    unless Sketchup::Image.method_defined?(:faceme)
      def faceme()
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return false
        end#if
        type=self.get_attribute("Animation", "faceme", nil)
        self.set_attribute("Animation", "facemeID", nil) if not type
        return type
      end
    end#if
   
    unless Sketchup::Image.method_defined?(:start_animation)
      def start_animation()
        self.definition.instances.each{|ins|
            ins.stop_animation()
        }
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return 0
        end#if
        ### get folder
        folder=self.get_attribute("Animation", "folder", nil)
				###
				folder = SKMtools.encoder(folder)
				###
        if ! folder
            puts "Image has no 'folder' set."
            return 0
        elsif ! File.exist?(folder)
            puts "Image 'folder' was not found: reset to 'nil'."
            self.set_attribute("Animation", "folder", nil)
            self.set_attribute("Animation", "timer", nil)
            return 0
        end#if
        model=self.model
        ### get timer
        timer=self.get_attribute("Animation", "timer", nil)
        ###
        if ! timer ### aniamte at scene change
          $imageAnimationObservers=[] if ! $imageAnimationObservers
          obs=ImageAnimationPagesObserver.new(self, folder)
          model.pages.add_observer(obs)
          $imageAnimationObservers << [self, obs]
          $imageAnimationObservers.uniq!
        else ### animate within scene
          timer=timer.to_f
          imgs=[".jpg",".jpeg",".png",".psd",".tif",".tga",".bmp"]
          imagefiles=[]
          Dir.entries(folder).each{|img|
            next if ! imgs.include?(File.extname(img).downcase)
            imagefiles << File.join(folder, img)
          }
          imagefiles.sort!
          self.imagefile=imagefiles[0]
          i=1
          tot=imagefiles.length
          tid=UI.start_timer(timer, true){
              self.imagefile=imagefiles[i]
              i+=1
              i=0 if i >= tot
          }
          $imageAnimationObservers=[] if ! $imageAnimationObservers
          $imageAnimationObservers << [self, tid]
          $imageAnimationObservers.uniq!
        end#if
        return 1
      end
    end#if
    
   unless Sketchup::Image.method_defined?(:start_animation_all)
    def start_animation_all()
      model=self.model
      num=0
      images=[]
      model.definitions.each{|d|images << d if d.image?}
      images.each{|image|
        doneone=false
        image.instances.each{|inst|
          folder=inst.get_attribute("Animation", "folder", nil)
					###
					folder = SKMtools.encoder(folder)
					###
          next if ! folder
          if ! File.exist?(folder)
            inst.set_attribute("Animation", "folder", nil)
            next
          end#if
          next if doneone
          ### get timer
          timer=inst.get_attribute("Animation", "timer", nil)
          ###
          inst.definition.instances.each{|ins|
              ins.stop_animation()
          }
          ###
          if ! timer ### animate at scene change
            $imageAnimationObservers=[] if ! $imageAnimationObservers
            obs=ImageAnimationPagesObserver.new(inst, folder)
            model.pages.add_observer(obs)
            $imageAnimationObservers << [inst, obs]
            $imageAnimationObservers.uniq!
          else ### animate within scene
            timer=timer.to_f
            imgs=[".jpg",".jpeg",".png",".psd",".tif",".tga",".bmp"]
            imagefiles=[]
            Dir.entries(folder).each{|img|
              next if ! imgs.include?(File.extname(img).downcase)
              imagefiles << File.join(folder, img)
            }
            imagefiles.sort!
            inst.imagefile=imagefiles[0]
            i=1
            tot=imagefiles.length
            tid=UI.start_timer(timer, true){
                inst.imagefile=imagefiles[i]
                i+=1
                i=0 if i >= tot
            }
            $imageAnimationObservers=[] if ! $imageAnimationObservers
            $imageAnimationObservers << [inst, tid]
            $imageAnimationObservers.uniq!
          end#if
          doneone=true ### only start one instance animating
          num+=1
        }
      }
      return num
    end
   end#if
    
   unless Sketchup::Image.method_defined?(:stop_animation)
    def stop_animation()
        if ! self.valid?
            puts "Image #{self} is no longer valid."
            return 0
        end#if
        return 0 if ! $imageAnimationObservers
        entry=nil
        $imageAnimationObservers.each{|a|
          if self == a[0] 
            entry=a
            break
          end#if
        }
        return 0 if not entry
        model=self.model
        begin
          if entry[1].class==Fixnum
            UI.stop_timer(entry[1])
          else
            model.pages.remove_observer(entry[1])
          end#if
          $imageAnimationObservers = $imageAnimationObservers-[entry]
          return 1
        rescue
          return -1
        end
        return 0
    end
   end#if
    
   unless Sketchup::Image.method_defined?(:stop_animation_all)
    def stop_animation_all()
        model=self.model
        num=0
        $imageAnimationObservers.each{|a|
          if not a[0].valid?
            $imageAnimationObservers = $imageAnimationObservers-[a]
            next
          end#if
          begin
            if a[1].class==Fixnum
              UI.stop_timer(a[1])
            else
              a[0].model.pages.remove_observer(a[1])
            end#if
            $imageAnimationObservers = $imageAnimationObservers-[a]
            num+=1
          rescue
            ###
          end
        }
        return num
    end
   end#if
=begin    
   unless Sketchup::Image.method_defined?(:xflipped?)
    def xflipped?()
        trans=self.transformation.to_a
        xscale = Math.sqrt(trans[0]**2+trans[1]**2+trans[2]**2)
        face=nil
        self.definition.entities.each{|e|
          if e.class==Sketchup::Face
            face=e
            break
          end#if
        }
        bb=face.bounds###self.bounds
        cen=bb.center
        #p min=bb.min
        p min=face.vertices[0].position
        zro=self.zrotation
        p org=self.origin
        nor=self.normal
        tr=Geom::Transformation.rotation(cen,nor,-zro)
        min.transform!(tr * self.transformation)
        xscale = -xscale if trans[0] > 0 and org != min
        if xscale > 0
          return false
        elsif xscale < 0
          return true
        else
          return nil
        end#if
    end#xscale
   end#if
    
   unless Sketchup::Image.method_defined?(:yflipped?)
    def yflipped?()
        trans=self.transformation.to_a
        yscale = Math.sqrt(trans[4]**2+trans[5]**2+trans[6]**2)
        face=nil
        self.definition.entities.each{|e|
          if e.class==Sketchup::Face
            face=e
            break
          end#if
        }
        bb=face.bounds###self.bounds
        cen=bb.center
        p 11111111111111111
        #p min=bb.min
        p min=face.vertices[0].position
        #p max=bb.max
        p max=face.vertices[2].position
        zro=self.zrotation
        p org=self.origin
        nor=self.normal
        tr=Geom::Transformation.rotation(cen,nor,-zro)
        p min.transform!(tr * self.transformation)
        p max.transform!(tr * self.transformation)
        yscale = -yscale if trans[5] < 0 and org != max and org != min
        if yscale > 0
          return false
        elsif yscale < 0
          return true
        else
          return nil
        end#if
    end#yscale i=Sketchup.active_model.selection[0];p i.xflipped?; p i.yflipped?
   end#if
=end
    
end#class
###########################################################################
### Pages Observer
class ImageAnimationPagesObserver < Sketchup::PagesObserver
    @@imgs=[".jpg",".jpeg",".png",".psd",".tif",".tga",".bmp"]
    def initialize(image, folder)
        @image=image
				###
				folder = SKMtools.encoder(folder)
				###
        @imagefiles=[]
        Dir.entries(folder).sort.each{|img|
          next if ! @@imgs.include?(File.extname(img).downcase)
          @imagefiles << File.join(folder, img)
        }
        @tot=@imagefiles.length
        @i=0
        @image.imagefile=@imagefiles[@i]
    end
    def onContentsModified(pages)
        @i+=1
        @i=0 if @i >= @tot
        @image.imagefile=@imagefiles[@i]
    end
end#class
###########################################################################
### Remove observers on close/quit etc with an AppObserver
class ImageAnimationAppObserver < Sketchup::AppObserver
  def onNewModel(model)  ### new from menu or the app
    if $imageAnimationObservers###
      $imageAnimationObservers[0][0].stop_animation_all()
    end
  end
  def onOpenModel(model) ### open other SKP from menu
    if $imageAnimationObservers###
      $imageAnimationObservers[0][0].stop_animation_all()
    end
  end
  def onQuit() ### users quits SUp completely [=MAC safe?]
    if $imageAnimationObservers###
      $imageAnimationObservers[0][0].stop_animation_all()
    end
  end
end
# Attach the observer at kick-off
Sketchup.add_observer(ImageAnimationAppObserver.new)
###########################################################################
