=begin
TIG (c) 2010 = 2013
  All Rights Reserved.
  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
  WARRANTIES,INCLUDING,WITHOUT LIMITATION,THE IMPLIED WARRANTIES OF 
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
###########################################################################
  material_class.rb
###########################################################################
Loaded by SKMtools_loader.rb in the ../Plugins/ folder.
Adds new methods to to the Material class:
###########################################################################
  material.save_texture('folder')
It returns the 'folder/texture.jpg' [or .png etc] if successful or 'nil'.
 e.g. texture=face.material.save_texture("C:\\Temp\\")
 >>> C:\Temp\Brick99.jpg
If the 'folder' does not exist there is an error message.
###########################################################################
  material.save_thumbnail('filepath')
  [or its alias material.write_thumbnail('filepath')]
It returns the 'filepath.png' if successful or 'nil' if not.
 e.g. thumb=face.material.save_thumbnail("C:\\Temp\\mythumb.png")
 >>> C:\Temp\mythumb.png
[If there's not a '.png' file-type suffix it will be made as one anyway!]
If the specified 'filepath' is a 'folder' then the thumbnail is made there 
named 'doc_thumbnail.png' [the default name]
 e.g. thumb=face.material.save_thumbnail("C:\\Temp")
 >>> C:\Temp\doc_thumbnail.png
The thumbnail resolution is always 64x64px*** - always as a '.png' file.
If you specify a resolution it reverts to 64 anyway!
If the 'folder' does not exist there is an error message.
Needs Thumbnailer.class and ThumbnailerRGB.class in the same sub-folder.
If you are in SUp >=v8-M1 the built in method is used instead
 'material.write_thumbnail(path, resol)' 
- note, that the built-in method returns 'true' if successful AND requires 
a alternative resolution to be passes as the second argument***, also you 
must specify the full path+name and not just a folder path.
###########################################################################
  material.save_bumpmap('folder', boolean)
It makes the 'folder/texture.jpg' [or .png etc] and also a 'texturebumpmap'.
Returns 'folder/texturebumpmap.png' [always .png] if successful, or 'nil'.
The 2nd argument 'boolean' is optional - it defaults to 'false' if not set.
  e.g. face.material.save_bumpmap("C:\\Temp\\")
  >>> C:\Temp\Brick99_BMP.png
If you give it the second argument as 'true' it 'inverts' the Bumpmap,
  e.g. invert=true; face.material.save_bumpmap("C:\\Temp\\", invert)
  >>> C:\Temp\Brick99_BMPI.png [with 'inverted' Bump colors]
Choose to invert if you want to swap light and dark values...
If the specified 'folder' does not exist there is an error message.
Needs 'Bumpmapper.jar' & 'BumpmapperInverted.jar' in its folder to run.
###########################################################################
  material.export_skm(folder)
It returns the 'folder/matname.skm' if successful or 'nil'.
  e.g. thumb=face.material.save_texture("C:\\Temp\\")
  >>> C:\Temp\Brick99.skm
If the 'folder' does not exist there is an error message.
Requires SKM_class [SM.export(mat,'folder')] loaded too.
###########################################################################
  material.delete [or its alias material.remove]
It deletes that Material from the Model and any Entities that used it get 
the Default Material [nil] instead...
Returns true/false on success/failure.
If you are in SUp >=v8-M1 the built in method 'material.remove' is used instead.
###########################################################################
  material.name="new_name"
it makes a cloned new-material from the original-material and gives it the 
'new_name', if the name 'new_name' exists then it's made unique with a 
numerical suffix, it gives the new-material to any entities that had the 
original-material, the original-material is then deleted from the model 
and material-browser: if the current-material in the browser was the 
original-material then the current-material in the browser becomes this 
new_material...  It returns the new 'Material' name: to keep a reference 
to the 'replaced' material use:
 nname=(omat.name='newname'); omat=model.materials[nname] if nname
If you are in SUp >=v8-M1 the built in method 'material.name=' is used instead.
###########################################################################
  material.save_texture_and_thumbnail('folder')
It returns the 'sub-folder' if successful or 'nil'.
e.g. model.materials["My Mat"].save_texture_and_thumbnail("C:\\Temp\\XXX")
>>> C:\Temp\XXX\My_Mat
If the 'folder' does not exist there is an error message.
The 'sub-folder' is named after the Material [with some characters made 
into '_'].  The 'sub-folder' contains the Material's 'texture file' - e.g. 
'Brick666.jpg' AND the equivalent thumbnail 'doc_thumbnail.png' - if the 
Material has NO texture then there is only the 'doc_thumbnail.png' file in 
the 'sub-folder'.
###########################################################################

###########################################################################
Donations:
  By Paypal.com to info @ revitrev.org
###########################################################################   
Version:
1.0 20101210 First issue as combined class script
1.1 20101212 New save_texture_and_thumbnail('folder') method added.
1.2 20101222 Glitch with MAC file paths addressed.
1.3 20110106 Robustness improved with existing method checks...
1.4 20110116 Aliases added for .write_thumbnail() and .remove.
             The .name='' and .delete methods now match SUp >=v8-M1 values.
1.5 20110116 Slightly recast 'aliases'.
1.6 20110213 MAC jar execution addressed.
1.7 20130827 Temp folder issues improved.
###########################################################################
=end
###
require('sketchup.rb')
###
class Sketchup::Material

unless Sketchup::Material.method_defined?(:save_texture)
def save_texture(path="")
  if ! self.valid?
    puts "Material #{self} is no longer valid."
    return false
  end#if
  return nil if ! self.texture #########################################
  if ! path.class==String
    path=""
  end#if
  if path==""
    puts "Usage: material.save_texture(folder)"
    return nil
  end#if
	###
	path = SKMtools.encoder(path)
	###
  if ! File.exist?(path)
    puts "Usage: material.save_texture(folder)"
    puts "Folder '#{path}' was not found."
    return nil
  end#if
  if ! File.ftype(path)=="directory"
    puts "Usage: material.save_texture(folder)"
    puts "Folder '#{path}' is not a 'Directory'."
    return nil
  end#if
  ###
  model=Sketchup.active_model
  ###
  pc=false
  pc=true if RUBY_PLATFORM =~ /mswin|mingw/
  ###
  tw=Sketchup.create_texture_writer
  ###
  temp_group=model.active_entities.add_group()
  temp_group.material=self
  iname=File.basename(self.texture.filename, ".*").gsub(/[^A-Za-z0-9_-]/, "_") + File.extname(self.texture.filename).downcase
  ### in case there a folder path
  ipath=File.join(path, iname)
  tw.load(temp_group)
  tw.write(temp_group, ipath)
  temp_group.erase! if temp_group.valid?
  ipath.tr!("/","\\") if pc
  ###
  if File.exist?(ipath)
    return ipath
  else
    return nil
  end#if
  
end#def
end#if

unless Sketchup::Material.method_defined?(:save_thumbnail)
def save_thumbnail(path="", resol=nil)
  if ! self.valid?
    puts "Material #{self} is no longer valid."
    return false
  end#if
  if ! path.class==String
    path=""
  end#if
  if path==""
    puts "Usage: material.save_thumbnail(filepath)"
    return nil
  end#if
  if File.exist?(path) and File.ftype(path)=="directory"
    path=File.join(path, "doc_thumbnail.png")
  end#if
  if ! File.exist?(File.dirname(path))
    puts "Usage: material.save_thumbnail(filepath)"
    puts "Folder '#{File.dirname(path)}' was not found."
    return nil
  end#if
  puts 'Resolution = 64' if resol
  if ! File.extname(path).downcase==".png"
    path=File.join(File.dirname(path), File.basename(path, ".*")+".png")
  end#if
  File.delete(path) if File.exist?(path)
  ###
  pc=false
  pc=true if RUBY_PLATFORM =~ /mswin|mingw/
  ###
  fol=SKMtools::FOLDER
  dir=SKMtools::TEMP
  begin
	Dir.mkdir(dir)
  rescue
  end
  model=Sketchup.active_model
  ###
  fromini=File.join(dir, "from.ini")
  toini=File.join(dir, "to.ini")
  rini=File.join(dir, "r.ini")
  gini=File.join(dir, "g.ini")
  bini=File.join(dir, "b.ini")
  ###
  path.tr!("/","\\") if pc
  file=File.new(toini, "w")
    file.write(path)
  file.close
  ###
  if self.texture
    ###
    texture_name=File.basename(self.texture.filename)### in case there a folder path
    frompath=File.join(dir, texture_name)
    originally_there=false
    originally_there=true if File.exist?(frompath)
    frompath.tr!("/","\\") if pc
    frompath=self.save_texture(dir)
    if ! frompath
      puts "Error: material.save_thumbnail"
      puts "Material's Texture file '#{texture_name}' was not made."
      return nil
    end#if
    file=File.new(fromini, "w")
      file.write(frompath)
    file.close
    jar=File.join(fol, "Thumbnailer.jar").tr("\\","/")
	jartemp=File.join(dir, "Thumbnailer.jar").tr("\\","/")
	  ###
	  file=File.open(jar, "rb")
		file.binmode
		data=file.read
	  file.close
	  ###
	  file=File.open(jartemp, "wb")
		file.binmode
		file.write(data)
	  file.close
	  ###
	  jar=jartemp
    ###
  else ### has a texture - use its image
    ###
    r=self.color.red.to_s
    g=self.color.green.to_s
    b=self.color.blue.to_s
    ###
    file=File.new(rini, "w")
      file.write(r)
    file.close
    file=File.new(gini, "w")
      file.write(g)
    file.close
    file=File.new(bini, "w")
      file.write(b)
    file.close
    jar=File.join(fol, "ThumbnailerRGB.jar").tr("\\","/")
	jartemp=File.join(dir, "ThumbnailerRGB.jar").tr("\\","/")
	  ###
	  file=File.open(jar, "rb")
		file.binmode
		data=file.read
	  file.close
	  ###
	  file=File.open(jartemp, "wb")
		file.binmode
		file.write(data)
	  file.close
	  ###
	  jar=jartemp
    ###
  end#if
  ###
  pwd=Dir.pwd
  Dir.chdir(dir)
    File.chmod(0777, jar) unless pc
    if jar[0].chr=="/"
        jar="file:#{jar}"
    else
        jar="file:/#{jar}"
    end#if
    UI.openURL(jar)###########################
  Dir.chdir(pwd)
  ###
  wait=0.0
  until File.exist?(path) || wait>=2.0
    sleep(0.1)### wait till made
    wait+=0.1
  end
  ###
  File.delete(frompath) if ! originally_there && frompath && File.exist?(frompath)
  ### remove temp ini files
  File.delete(fromini) if File.exist?(fromini)
  File.delete(toini) if File.exist?(toini)
  File.delete(rini) if File.exist?(rini)
  File.delete(gini) if File.exist?(gini)
  File.delete(bini) if File.exist?(bini)
  ###
  if path && File.exist?(path)
    return path
  else
    return nil
  end#if
  ### 
end#def
end#if

unless Sketchup::Material.method_defined?(:save_bumpmap)
def save_bumpmap(path="",inverted=false)
  if ! self.valid?
    puts "Material #{self} is no longer valid."
    return false
  end#if
  return nil if ! self.texture
  if ! path.class==String
    path=""
  end#if
  if path==""
    puts "Usage: material.save_texture(folder)"
    return nil
  end#if
  if ! File.exist?(path)
    puts "Usage: material.save_texture(folder)"
    puts "Folder '#{path}' was not found."
    return nil
  end#if
  if ! File.ftype(path)=="directory"
    puts "Usage: material.save_texture(folder)"
    puts "Folder '#{path}' is not a 'Directory'."
    return nil
  end#if
  ###
  model=Sketchup.active_model
  ###
  fol=SKMtools::FOLDER
  dir=SKMtools::TEMP
  begin
	Dir.mkdir(dir)
  rescue
  end
  ###
  pc=false
  pc=true if RUBY_PLATFORM =~ /mswin|mingw/
  ###
  frompath=self.save_texture(path)
  frompath.tr!("/","\\") if pc
  if ! frompath
    puts "Error: material.save_texture_bumpmap"
    puts "Material's Texture file '#{texture_name}' was not made."
    return nil
  end#if
  ###
  fromini=File.join(dir, "from.ini")
  toini=File.join(dir, "to.ini")
  idir=File.dirname(frompath)
  ibase=File.basename(frompath, ".*")
  #iext=File.extname(frompath)
  iext=".png"
  bmp="_BMP"
  bmp=bmp+"I" if inverted
  topath=File.join(idir, ibase+bmp+iext)
  topath.tr!("/","\\") if pc
  file=File.new(fromini, "w")
    file.write(frompath)
  file.close
  file=File.new(toini, "w")
    file.write(topath)
  file.close
  ###
  jar=File.join(fol, "Bumpmapper.jar").tr("\\","/")
  jar=File.join(fol, "BumpmapperInverted.jar").tr("\\","/") if inverted
  jartemp=File.join(dir, "Bumpmapper.jar").tr("\\","/")
  jartemp=File.join(dir, "BumpmapperInverted.jar").tr("\\","/") if inverted
  ###
  file=File.open(jar, "rb")
	file.binmode
	data=file.read
  file.close
  ###
  file=File.open(jartemp, "wb")
	file.binmode
	file.write(data)
  file.close
  ###
  jar=jartemp
  ###
  pwd=Dir.pwd
  Dir.chdir(dir)
    File.chmod(0777, jar) unless pc
    if jar[0].chr=="/"
        jar="file:#{jar}"
    else
        jar="file:/#{jar}"
    end#if
    UI.openURL(jar)###########################
  Dir.chdir(pwd)
  ###
  wait=0.0
  until File.exist?(topath) || wait>=2.0
    sleep(0.1)### wait till made
    wait+=0.1
  end
  ###
  File.delete(fromini) if File.exist?(fromini)
  File.delete(toini) if File.exist?(toini)
  if File.exist?(topath)
    return topath
  else
    return nil
  end#if
  
end#def
end#if

unless Sketchup::Material.method_defined?(:export_skm)
def export_skm(folder="")
  if ! self.valid?
    puts "Material #{self} is no longer valid."
    return false
  end#if
  if ! folder.class==String
    folder=""
  end#if
  if folder==""
    puts "Usage: material.export_skm(folder)"
    return nil
  end#if
	###
	folder = SKMtools.encoder(folder)
	###
  if ! File.exist?(folder)
    puts "Usage: material.export_skm(folder)"
    puts "Folder '#{folder}' was not found."
    return nil
  end#if
  if ! File.ftype(folder)=="directory"
    puts "Usage: material.export_skm(folder)"
    puts "Folder '#{folder}' is not a 'Directory'."
    return nil
  end#if
  ###
  model=Sketchup.active_model
  ###
  ipath=SKM.export(self, folder)
  ###
  if File.exist?(ipath)
    return ipath
  else
    return nil
  end#if
  
end#def
end#if

unless Sketchup::Material.method_defined?(:delete)
def delete()
    if ! self.valid?
      puts "Material #{self} is no longer valid."
      return false
    end#if
    model=Sketchup.active_model
    ents=model.entities
    defs=model.definitions
    mats=model.materials
    ents.each{|e|
      e.material=nil if e.respond_to?(:material) && e.material==self
      e.back_material=nil if e.kind_of?(Sketchup::Face) && e.back_material==self
    }
    defs.each{|d|
      d.entities.each{|e|
        e.material=nil if e.respond_to?(:material) && e.material==self
        e.back_material=nil if e.kind_of?(Sketchup::Face) && e.back_material==self
      }
    }
    group=ents.add_group()
    gents=group.entities ### temporarily use other materials
    temp=gents.add_group()
    (mats.to_a-[self]).each{|m|
      tc=temp.copy
      tc.material=m
    }
    mats.current=nil if mats.current==self ### ensure self is not current material
    mats.purge_unused ### purge self from browser
    group.erase! if group.valid? ### erase! the temporary material user, use set as was.
    return true
end#def
end#if

unless Sketchup::Material.method_defined?(:name=)
def name=(new_name=nil)
    if ! self.valid?
      puts "Material #{self} is no longer valid."
      return nil
    end#if
    if ! new_name || ! new_name.class==String || new_name=="" || new_name==" "
      puts "Material 'name' must be a String."
      return self
    end#if
    if new_name==self.display_name
      puts "The Material 'name' ia already = '#{new_name}'."
      return self
    end#if
    model=Sketchup.active_model
    ents=model.entities
    defs=model.definitions
    mats=model.materials
    num=1
    while mats[new_name]
      new_name=new_name + num.to_s
      num+=1
    end#while ### so it gets a unique name
    ### get self details
    alpha=self.alpha
    color=self.color
    texture=self.texture
    attrdicts=self.attribute_dictionaries ###################
    ### make new material to match self
    newMat=mats.add(new_name)
    newMat.alpha=alpha
    newMat.color=color
    ###################
    attrdicts.each{|attrdict|
      name=attrdict.name
      attrdict.each_pair{|key,value|newMat.set_attribute(name,key,value)}
    } if attrdicts
    ###################
    if texture
      tdir=SKMtools::TEMP
      filename=File.basename(texture.filename)      
      filepath=File.join(tdir, filename)
      group=ents.add_group
      group.material=self
      tw=Sketchup.create_texture_writer
      tw.load(group)
      tw.write(group,filepath)
      group.erase! if group.valid?
      newMat.texture=filepath if File.exist?(filepath)
      File.delete(filepath) if File.exist?(filepath)
      size=[texture.width, texture.height]
      newMat.texture.size=size if newMat.texture
    end#if
    ents.each{|e| ### now tidy up the materials
      e.material=newMat if e.respond_to?(:material) && e.material==self
      e.back_material=newMat if e.kind_of?(Sketchup::Face) && e.back_material==self
    }
    defs.each{|d|
      d.entities.each{|e|
        e.material=newMat if e.respond_to?(:material) && e.material==self
        e.back_material=newMat if e.kind_of?(Sketchup::Face) && e.back_material==self
      }
    }
    group=ents.add_group()
    gents=group.entities ### temporarily use all other materials
    temp=gents.add_group()
    (mats.to_a-[self]).each{|m|
      tc=temp.copy
      tc.material=m
    }
    mats.current=newMat if mats.current==self ### ensure self is NOT current material
    mats.purge_unused ### purge 'self' from browser
    group.erase! if group.valid? ### erase! the temporary material user, use set as was.
    ###
    return newMat
    ###
end#def
end#if

unless Sketchup::Material.method_defined?(:save_texture_and_thumbnail)
def save_texture_and_thumbnail(path="")
  if ! self.valid?
    puts "Material #{self} is no longer valid."
    return false
  end#if
  if ! path.class==String
    path=""
  end#if
  if path==""
    puts "Usage: material.save_texture(folder)"
    return nil
  end#if
	###
	folder = SKMtools.encoder(folder)
	###
  if ! File.exist?(path)
    puts "Usage: material.save_texture(folder)"
    puts "Folder '#{path}' was not found."
    return nil
  end#if
  if ! File.ftype(path)=="directory"
    puts "Usage: material.save_texture(folder)"
    puts "Folder '#{path}' is not a 'Directory'."
    return nil
  end#if
  ###
  pc=false
  pc=true if RUBY_PLATFORM =~ /mswin|mingw/
  ###
  name=self.display_name.gsub(/[^A-Za-z0-9_-]/, "_")
  tdir=File.join(path, name)
  tdir.tr!("/","\\") if pc
  begin
	Dir.mkdir(tdir)
  rescue
  end
  ###
  model=Sketchup.active_model
  ###
  ###
  if self.texture
    tw=Sketchup.create_texture_writer
    ###
    temp_group=model.active_entities.add_group()
    temp_group.material=self
    iname=File.basename(self.texture.filename, ".*").gsub(/[^A-Za-z0-9_-]/, "_") + File.extname(self.texture.filename).downcase
    ### in case there a folder
    ipath=File.join(tdir, iname)
    tw.load(temp_group)
    tw.write(temp_group, ipath)
    temp_group.erase! if temp_group.valid?
    if ! File.exist?(ipath)
      puts "Error: material.save_texture_and_thumbnail"
      puts "Material's Texture file '#{iname}' was not made."
      return nil
    end#if
  end#if
  ### now make thumbnail
  fol=SKMtools::FOLDER
  dir=SKMtools::TEMP
  begin
	Dir.mkdir(dir)
  rescue
  end
  fromini=File.join(dir, "from.ini")
  toini=File.join(dir, "to.ini")
  rini=File.join(dir, "r.ini")
  gini=File.join(dir, "g.ini")
  bini=File.join(dir, "b.ini")
  ###
  tpath=File.join(tdir, "doc_thumbnail.png")
  ### toini
  tpath.tr!("/","\\") if pc
  file=File.new(toini, "w")
    file.write(tpath)
  file.close
  ###
  if self.texture ### make thumbnail from ipath
    file=File.new(fromini, "w")
      file.write(ipath)
    file.close
    jar=File.join(fol, "Thumbnailer.jar").tr("\\","/")
	jartemp=File.join(dir, "Thumbnailer.jar").tr("\\","/")
	  ###
	  file=File.open(jar, "rb")
		file.binmode
		data=file.read
	  file.close
	  ###
	  file=File.open(jartemp, "wb")
		file.binmode
		file.write(data)
	  file.close
	  ###
	  jar=jartemp
    ###
  else ### make thumbnail from RGB
    ###
    r=self.color.red.to_s
    g=self.color.green.to_s
    b=self.color.blue.to_s
    ###
    file=File.new(rini, "w")
      file.write(r)
    file.close
    file=File.new(gini, "w")
      file.write(g)
    file.close
    file=File.new(bini, "w")
      file.write(b)
    file.close
    jar=File.join(dir, "ThumbnailerRGB.jar").tr("\\","/")
	jartemp=File.join(dir, "ThumbnailerRGB.jar").tr("\\","/")
	  ###
	  file=File.open(jar, "rb")
		file.binmode
		data=file.read
	  file.close
	  ###
	  file=File.open(jartemp, "wb")
		file.binmode
		file.write(data)
	  file.close
	  ###
	  jar=jartemp
    ###
  end#if
  ###
  pwd=Dir.pwd
  Dir.chdir(dir)
    File.chmod(0777, jar) unless pc
    if jar[0].chr=="/"
        jar="file:#{jar}"
    else
        jar="file:/#{jar}"
    end#if
    UI.openURL(jar)###########################
  Dir.chdir(pwd)
  ###
  wait=0.0
  until File.exist?(tpath) || wait>=2.0
    sleep(0.1)### wait till made
    wait+=0.1
  end
  sleep(0.1)
  if ! File.exist?(tpath)
    puts "Error: material.save_texture_and_thumbnail"
    puts "Unable to make 'doc_thumbnail.png'"
  end#if
  ### remove temp ini files
  File.delete(fromini) if File.exist?(fromini)
  File.delete(toini) if File.exist?(toini)
  File.delete(rini) if File.exist?(rini)
  File.delete(gini) if File.exist?(gini)
  File.delete(bini) if File.exist?(bini)
  ###
  if File.exist?(tdir)
    return tdir
  else
    return nil
  end#if
  ###
end#def
end#if

unless Sketchup::Material.method_defined?(:write_thumbnail)
 def write_thumbnail(path, res)
   self.save_thumbnail(path, res)
 end
end#if

unless Sketchup::Material.method_defined?(:remove)
 def remove()
   self.delete()
 end
end#if

end#class
###
