=begin
TIG (c) 2010 - 2013
  All Rights Reserved.
  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
  WARRANTIES,INCLUDING,WITHOUT LIMITATION,THE IMPLIED WARRANTIES OF 
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
###########################################################################
  podium_texture+bumpmap.rb
###########################################################################
Loaded by SKMtools_loader.rb in the ../Plugins/ folder.
Adds a new tool:
  podiumtexturebumpmap(materialname, folder, boolean)
It requires other scripts - material.export_skm.rb plus 
material.save_bumpmap.rb, material.save_bumpmapinverted.rb
and the associated .jar files for each of those scripts.
To run type
 podiumtexturebumpmap("Brick66","C:\\Temp\\MyPodiumSKMs")
It finds the Model's Material named "Brick 66" and saves a SKM of it to the 
specified folder [that must exist], with a bumpmap with the same name.
If the 3rd argument ['boolean'] is given 'true' the bumpmap is 'inverted'.
The SKM is name with a 'preffix' - 'PODIUM_Brick_66.skm' and the bumpmap as 
'PODIUM_Brick_66-bumpmap.png', the Material.texture IF any is also made 
named 'PODIUM_Brick_66.pjpg' [or .png as appropriate] - Note that this 
name might not match the texture.name inside the SKM itself at all...
It returns a three element array: 
 [path_to_material.skm, path_to_bumpmap.png, path_to_texture]
The 2nd and 3rd elements might be nil if there is no texture associated 
with the material [the bumpmap is little use on a 'flat' material!]
###########################################################################
Donations:
  By Paypal.com to info @ revitrev.org
###########################################################################   
Version:
1.0 20101206 First issue
1.1 20120827 Temp folder resolved.
###########################################################################
=end
###
require('sketchup.rb')
###

def podiumtexturebumpmap(mname="?", path="", inverted=false)
  ###
  if ! path.class==String
    path=""
  end#if
  if path==""
    puts "Usage: podiumtexturebumpmap(materialname, folder, boolean)"
    return nil
  end#if
	###
	path = SKMtools.encoder(path)
	###
  if ! File.exist?(path)
    puts "Usage: podiumtexturebumpmap(materialname, folder, boolean)"
    puts "Folder '#{path}' was not found."
    return nil
  end#if
  if ! File.ftype(path)=="directory"
    puts "Usage: podiumtexturebumpmap(materialname, folder, boolean)"
    puts "Folder '#{path}' is not a 'Directory'."
    return nil
  end#if
  ###
  model=Sketchup.active_model
  mats=model.materials
  if ! mat=mats[mname]
    puts "Usage: podiumtexturebumpmap(materialname, folder, boolean)"
    puts "Material '#{mname}' is not in the Model's Material List."
    return nil
  end#if
  ###
  pc=false
  pc=true if RUBY_PLATFORM =~ /mswin|mingw/
  ### save dir contents at start
  extg=Dir.entries(path).to_a
  ### make SKM
  skm=mat.export_skm(path)
  if ! skm || ! File.exist?(skm)
    puts "Usage: podiumtexturebumpmap(materialname, folder, boolean)"
    puts "SKM '#{mname}' could not be made."
    return nil
  end#if
  skmname=File.basename(skm, ".*")
  newskmname="PODIUM_"+skmname+".skm"
  newskm=File.join(path, newskmname)
  File.delete(newskm) if File.exist?(newskm)
  sleep(0.1)
  File.rename(skm, newskm)
  newskm.tr!("/","\\") if pc
  ### make bumpmap
  if inverted
    bump=mat.save_bumpmap(path, true)
  else
    bump=mat.save_bumpmap(path)
  end#if
  if bump
    newbumpname="PODIUM_"+skmname+"-bumpmap.png"
    newbump=File.join(path, newbumpname)
    File.delete(newbump) if File.exist?(newbump)
    sleep(0.1)
    File.rename(bump, newbump)
    newbump.tr!("/","\\") if pc
  else
    puts "Bumpmap file NOT made."
    newbump=nil ### no texture to make bumpmap from!
  end#if
  ### rename the extra texture file
  texname=((Dir.entries(path).to_a) - [newskmname, newbumpname] - extg)[0]
  if texname
    tex=File.join(path, texname)
    texext=File.extname(tex)
    newtexname="PODIUM_"+skmname+texext  
    newtex=File.join(path, newtexname)
    File.delete(newtex) if File.exist?(newtex)
    sleep(0.1)
    File.rename(tex, newtex)
    newtex.tr!("/","\\") if pc
  else
    puts "Texture file NOT made."
    newtex=nil ### no texture to make bumpmap from!
  end#if
  
  ###
  return [newskm, newbump, newtex]
  ###
end#def
###
