=begin
TIG (c) 2010 - 2013
  All Rights Reserved.
  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
  WARRANTIES,INCLUDING,WITHOUT LIMITATION,THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
###########################################################################
  skm_class.rb
###########################################################################
Needs 'ZipDirectory.jar' and 'UnZipDirectory.jar',
and 'Thumbnailer.jar' and 'ThumbnailerRGB.jar' in the same sub-folder.
You will need Java installed on your PC/MAC - it usually is...
Also needs 'material-save_texture.rb' to export a SKM's texture.
[Related files are 'material-save_thumbnail.rb' & 'material-export_skm.rb'.]
Temporary .ini files are written to user's Temp folder copied from the 
../Plugins/SKMtools/ folder & deleted by the various methods to pass values to 
the .jar executable.
Loaded by 'SKMtools_loader.rb' in the ../Plugins/ folder.
###########################################################################
Adds a new module 'SKM' - with methods to import, export, save_texture and
save_thumbnail from SKM files etc.
###########################################################################

SKM.import(path_skm)
    It returns an error if the file 'path_skm' doesn't exist or have a
    '.skm' suffix.
    It makes a temporary SKM folder with the extracted SKM files contents
    [overwriting any existing one], e.g.
        '../Materials/My Material.skm'
    makes a temporary SKM folder as
        '../Plugins/SKMtools/Temp/My_Material/'
    [note the SKM name using [A-Za-z0-9_-] characters only]
    The required values are read from the files and a new material is added
    to the model based on these, named.
        e.g. 'My_Material'
    If that material name exists then it increments with a numerical suffix,
        e.g. 'My_Material1'
    The temporary SKM folder/files are deleted [unloaded].
    Returns a new 'Sketchup::Material' if successful, or 'nil'.
    Note that any attribute-dictionaries in the SKM xml file ARE imported
    with the new Material.

SKM.export(material, folder)
    Takes the given 'material' [as a 'Sketchup::Material' Object] and the
    destination 'folder' to receive the exported SKM file [as a 'String'.]
    It return errors if 'material' or 'folder' do not exist.
    e.g.
    For the 'material'
        'My Material'
    with 'folder'
        '../Materials/'
    it gives the SKM file as
        '../Materials/My_Material.skm'
    [note the SKM name using [A-Za-z0-9_-] characters only]
    It makes a temporary SKM folder with the default SKM files in it called
        '../Plugins/SKMtools/Temp/My_Material/'
    The required values are read from the 'material' and added to these
    files in the temporary SKM folder.
    The temporary SKM folder is compressed into a SKM file in the
    destination 'folder'.
    If the SKM exists in that folder it is overwritten.
    The temporary SKM folder/files are deleted [unloaded] on completion.
    Returns the path to a new 'SKM' file if successful, or 'nil'.
    Any attribute-dictionaries in the 'material' ARE put into the SKM xml.

SKM.save_texture(path_skm, folder)
    If the file 'path_skm' or 'folder' doesn't exist there's an error.
    It makes a temporary SKM folder with the extracted SKM files
    [overwriting any existing one], e.g.
        '../Materials/My Material.skm'
    makes a temporary folder as
        '../Plugins/SKMtools/Temp/My_Material/'
    [using [A-Za-z0-9_-] characters in the name only]
    If the SKM has no 'texture' file it returns 'nil'.
    Otherwise the 'texture' file - e.g. 'My_Materials_1.jpg' - is moved to
    the destination 'folder'.
    The temporary SKM folder/files are deleted [unloaded].
    Returns the 'path to the texture file, or 'nil' if failed or the SKM
    has ho texture file.

SKM.save_thumbnail(path_skm, file_png)
    If the 'file' specified by 'file_png' is actually an existing
    'directory' then the thumbnail is extracted into that folder using its
    original name - 'doc_thumbnail.png'.
    If the file 'file_png' is doesn't have a '.png' suffix it is added.
    If the destination file 'file_png' exists it will be overwritten.
    If the file 'path_skm' or the 'parent-folder' for the 'file_png' file
    doesn't exist there's an error.
    It makes a temporary SKM folder with the extracted SKM files
    [overwriting any existing one], e.g.
        '../Materials/My Material.skm'
    makes a temporary folder as
        '../Plugins/SKMtools/Temp/My_Material/'
    [using [A-Za-z0-9_-] characters in the name only]
    If the SKM has no 'thumbnail' file it returns 'nil' - but this is very
    unlikely as every SKM should contain a file called 'doc_thumbnail.png'
    [64x64pxs].
    The 'thumbnail' file
    - e.g. 'My_Materials_thumbnail.jpg' or 'doc_thumbnail.png' -
    is moved to the specified destination 'parent-folder'.
    The temporary SKM folder/files are deleted [unloaded].
    Returns the 'path to the thumbnail .png file, or 'nil' if failed.
    
SKM.extract(path_skm, [folder])
    This extracts all of the data from the SKM file specifed in 'path_skm'.
    The SKM's data is returned as a 'hash' [see below for contents...]
    If the optional 'folder' is specified the extracted SKM data is 
    retained in a sub-folder named after the SKM file, e.g. 'MY Mat.skm' >>> 
    a new folder called 'My_Mat'. This sub-folder contains all of the SKM's 
    files as follows: The SKM's thumbnail called 'doc_thumbnail.png', 
    three files containing assorted data relating to the SKM [document.xml, 
    documentProperties.xml and references.xml], and if the SKM has a 
    'texture' then there is also the 'image-file' is in a sub-folder called 
    'ref', e.g. /ref/'MyMat.jpg'.  If there is a 'folder' specified then it 
    returns additional entries in the 'hash' [see below for contents...]
    The 'data' hash contains the following:
     :path >>> the path to the original SKM file, e.g. 'C:\Temp\My Mat.skm'.
     :name >>> the SKM's name as it will SKM.import(), e.g. 'My_Mat'.
     :color >>> the SKM material's color as an array [r, g, b] - e.g. [225, 0, 0].
     :alpha >>> the SKM material's alpha [transparency/opacity] - e.g. 0.65.
     :texture >>> the SKM material's texture name, e.g. 'Brick666.jpg' or 'nil' if no texture
     :xscale >>> the SKM material's texture's xscale or 'nil' if no texture
     :yscale >>> the SKM material's texture's yscale or 'nil' if no texture  
     :avgcolor >>> the SKM material's texture's average-color, e.g. '2837892', or 'nil' if no 'texture'
     :attributes >>> the SKM material's attributes as an array, e.g. [[dict,key,value], [dict,key.value],...] or [] if none.
    The following are all set to 'nil' if the optional 'folder' is not specified
     :extracted_path >>> the path to the extracted SKM 'data', e.g. 'C:\Temp\My_Mat'.
     :extracted_thumbnail >>> the extracted SKM's thumnail path, e.g. 'C:\Temp\My_Mat\doc_thumbnail.png'.
     :extracted_texture >>> the extracted SKM's texture's path, e.g. 'C:\Temp\My_Mat\Brick666.jpg', or 'nil' if no 'texture'.
    So set the SKM data thus: data=SKM.extract('C:\Temp\My Mat.skm', 'C:\Temp')
    Then to find the SKM's 'color' you use color=data[:color] >>> [255, 0, 0] 
    or it's 'alpha' you use alpha=data[:alpha] >>> 0.65 etc.

SKM.save_texture_and_thumbnail(path_skm, folder)
    It returns the 'sub-folder' if successful or 'nil'.
    e.g. SKM.save_texture("C:Temp\\My Mat.skm", "C:\\Temp\\XXX")
    >>> C:\Temp\XXX\My_Mat
    If the 'folder' does not exist there is an error message.
    The 'sub-folder' is named after the SKM [with some characters made 
    into '_'].  The 'sub-folder' contains the SKM's 'texture file' - e.g. 
    'Brick666.jpg' AND the equivalent thumbnail 'doc_thumbnail.png', if the 
    SKM has NO texture then there is only the 'doc_thumbnail.png' file in 
    the 'sub-folder'.
    It returns the 'sub-folder' - e.g. C:\Temp\XXX\My_Mat
    
###########################################################################
Donations:
  By Paypal.com to info @ revitrev.org
###########################################################################
Version:
1.0 20101129 First issue
1.1 20101202 Now uses .jar and .ini files.
1.2 20101210 General update.
1.3 20101212 SKM.extract added.
1.4 20101222 Glitch with MAC file paths addressed.
1.5 20101230 Attribute value.class case test simplified.
1.6 20110213 MAC jar execution addressed.
1.7 20130827 Temp folder issues improved.
###########################################################################
=end
###
require 'sketchup.rb'
###

module SKM

include SKMtools

###########################################################################
### set variables
@dir=SKMtools::FOLDER
@temp=SKMtools::TEMP
@default=File.join(FOLDER, "SKMDefaultXMLs")
if ! File.exist?(@default) || ! File.ftype(@default)=="directory"
  UI.messagebox("Folder #{@default} missing!")
end#if
if ! File.exist?(@temp)
  begin
	Dir.mkdir(@temp)
  rescue
  end
elsif ! File.ftype(@temp)=="directory"
  File.delete(@temp)
  begin
	Dir.mkdir(@temp)
  rescue
  end
end#if
###
@pc=false
@pc=true if RUBY_PLATFORM =~ /mswin|mingw/
###########################################################################

def SKM.import(path="")

  if ! path.class==String || path==""
    puts "Usage: SKM.import(path_skm)"
    return nil
  end#if
	###
	path = SKMtools.encoder(path)
	###
  if ! File.exist?(path)
    puts "Usage: SKM.import(path_skm)"
    puts "File ''#{path}'' not found."
    return nil
  end#if
  if ! File.extname(path).downcase==".skm"
    puts "Usage: SKM.import(path_skm)"
    puts "File ''#{path}'' is not a SKM file."
    return nil
  end#if
  ###
  model=Sketchup.active_model
  materials=model.materials
  ###
  oskm=File.basename(path)
  dname=File.basename(path, ".*")
  skm_dir=File.join(@temp, dname)
  SKM.unload(skm_dir)
  name=dname.gsub(/[^A-Za-z0-9_-]/, "_")
  ###
  tskm=File.join(@temp, oskm)
  File.delete(tskm) if File.exist?(tskm)
  ###
  file=File.open(path, "rb")
    file.binmode
    data=file.read
  file.close
  ###
  file=File.open(tskm, "wb")
    file.binmode
    file.write(data)
  file.close
  ### make ini
  fromini=File.join(@temp, "from.ini")
  ###
  tskm.tr!("/","\\") if @pc
  file=File.new(fromini, "w")
    file.write(tskm)
  file.close
  ###
  jar=File.join(@dir, "UnZipDirectory.jar").tr("\\","/")
  jartemp=File.join(@temp, "UnZipDirectory.jar").tr("\\","/")
  ###
  file=File.open(jar, "rb")
    file.binmode
    data=file.read
  file.close
  ###
  file=File.open(jartemp, "wb")
    file.binmode
    file.write(data)
  file.close
  ###
  jar=jartemp
  ###
  pwd=Dir.pwd
  Dir.chdir(@temp)
    File.chmod(0777, jar) unless @pc
    if jar[0].chr=="/"
        jar="file:#{jar}"
    else
        jar="file:/#{jar}"
    end#if
    UI.openURL(jar) ###########################
  Dir.chdir(pwd)
  ###
  wait=0.0
  until File.exist?(File.join(skm_dir, "document.xml")) || wait>=2.0
    sleep(0.1)### wait till xml is made
    wait+=0.1
  end#
  sleep(0.1)
  unless File.exist?(File.join(skm_dir, "document.xml"))
    puts "SKM.import: Could not extract the SKM's 'document.xml'."
    File.delete(fromini) if File.exist?(fromini)
    File.delete(tskm) if File.exist?(tskm)
    SKM.unload(skm_dir)
    return nil
  end#if
  ## now make new material 'mat'
  mat=nil
  mat=materials.add(name)
  name=mat.name
  ### get values from xml files
  file=File.open(File.join(skm_dir, "document.xml"))
    text=file.read
  file.close
  ### parse values - colors
  r=(text.split(" colorRed=\""))[1].split("\"")[0].to_i
  g=(text.split(" colorGreen=\""))[1].split("\"")[0].to_i
  b=(text.split(" colorBlue=\""))[1].split("\"")[0].to_i
  ### alpha
  if (text.split(" useTrans=\""))[1].split("\"")[0].to_i
    a=1.0
  else
    a=(text.split(" trans=\""))[1].split("\"")[0].to_f
  end#if
  ### path to image IF set
  if text.include?(" path=\"")
    tf=(text.split(" path=\""))[1].split("\"")[0]
    tfi=File.join(File.join(skm_dir, "ref"), tf)
    tfi.tr!("/","\\") if @pc
    w=(text.split(" xScale=\""))[1].split("\"")[0].to_f
    h=(text.split(" yScale=\""))[1].split("\"")[0].to_f
  else
    tfi=nil
  end#if
  ### set mat's values
  mat.color=[r,g,b]
  mat.alpha=a
  ### mat's image
  if tfi ### wait for file to get made by the jar
    wait=0.0
    until File.exist?(tfi) || wait>=2.0
      sleep(0.1)
      wait+=0.1
    end#
    sleep(0.1)
    unless File.exist?(File.join(tfi))
      puts "SKM.import: Could not extract the SKM's texture '#{tf}'."
      File.delete(fromini) if File.exist?(fromini)
      File.delete(tskm) if File.exist?(tskm)
      SKM.unload(skm_dir)
      return nil
    end#if
    mat.texture=tfi
    mat.texture.size=[w,h] if mat.texture
  end#if
  ###
  ### sort out attributes #################################################
  if text.include?("<n0:AttributeDictionaries count=\"")
    attdics_num=text.split("<n0:AttributeDictionaries count=\"")[1].split("\"")[0].to_i
    attdics_data=text.split("<n0:AttributeDictionary count=\"")[1..-1]
    attdics=[]
    attdics_data.each{|dic|
      name=dic.split("\"")[2]
      kvs=[]
      dic.each{|d|
        d.split("<n0:Attribute key=\"")[1..-1].each{|kv|
          key=kv.split("\"")[0]
          type=kv.split(" type=\"")[1].split("\"")[0]
          if kv.include?("</n0:Attribute>")
            value=kv.split(">")[1].split("</n0:Attribute")[0]
          else
            value=nil
          end#if
          kvs << [key, type, value]
        }
      }
      attdics << [name, kvs]
    }
    ### add attributes
    attdics.each{|ad|
      name=ad[0]
      kvs=ad[1]
      kvs.each{|kv|
        key=kv[0]
        type=kv[1]
        value=kv[2]
        case type
          when "0"
            value=nil
          when "4"
            value=value.to_i
          when "6"
            value=value.to_f
          when "7"
            if value=="1"
              value=true
            else
              value=false
            end#if
          when "10"
            value=value
          when "11"
            value=[]
          else
            value=nil
        end#case
        mat.set_attribute(name, key, value)
      }
    }
  end#if
  ### #####################################################################
  ### tidy inis
  File.delete(fromini) if File.exist?(fromini)
  File.delete(tskm) if File.exist?(tskm)
  SKM.unload(skm_dir)
  ###
  if mat
    return mat
  else
    return nil
  end#if
  ###

end#import


def SKM.export(mat=nil, folder="")

  if ! mat || ! mat.class==Sketchup::Material
    puts "Usage: SKM.export(material, folder)"
    return nil
  end#if
  if ! mat.class==Sketchup::Material
    puts "Usage: SKM.export(material, folder)"
    puts "The specified 'material' is not a 'Sketchup::Material'."
    return nil
  end#if
  if mat.deleted?
    puts "Usage: SKM.export(material, folder)"
    puts "The specified 'material' is not 'valid'."
    return nil
  end#if
  if ! folder.class==String || folder==""
    puts "Usage: SKM.export(material, folder)"
    puts "The specified 'folder' is not a 'String'."
    return nil
  end#if
  if ! File.exist?(folder)
    puts "Usage: SKM.export(material, folder)"
    puts "Folder '#{folder}' was not found."
    return nil
  end#if
  if ! File.ftype(folder)=="directory"
    puts "Usage: SKM.export(material, folder)"
    puts "Folder '#{folder}' is not a 'Directory'."
    return nil
  end#if
  ###
  model=Sketchup.active_model
  ###
  tw=Sketchup.create_texture_writer
  ### get colors
  colorRed=mat.color.red.to_s
  colorGreen=mat.color.green.to_s
  colorBlue=mat.color.blue.to_s
  ### get alpha 1=opaque, 0=transparent
  trans=mat.alpha.to_s
  if mat.alpha==1.0
    useTrans="0"
  else
    useTrans="1"
  end#if
  ### colorize
  if mat.materialType < 2
    colorizeType="0"
  else
    colorizeType="1"
  end#if
  type=mat.materialType.to_s ### 0/1/2
  name=mat.display_name.gsub(/[^A-Za-z0-9_-]/, "_")
  title=mat.name.gsub(/[^A-Za-z0-9_-]/, "_")
  ###
  ### do attributes
  atts=[]
  attdics=mat.attribute_dictionaries
  attdics.each{|attdic|
    attdic.each_pair{|k,v|
      atts << [attdic.name, k, v] if v
    } if attdic
  } if attdics
  ### atts is an array of the material's attributes
  ### date
  created=Time.now.strftime("%Y-%m-%dT%H:%M:%S")
  modified=created
  ###
  ### ref to SKM
  skm=name+".skm"
  path=File.join(folder, skm)
  ### make temp SKM folder
  skm_dir=File.join(@temp, name)
  SKM.unload(skm_dir) ### fails
  Dir.mkdir(skm_dir)
  ### copy files around
  def_resXML=File.join(@default, "references.xml")
  def_docXML=File.join(@default, "document.xml")
  def_docPXML=File.join(@default, "documentProperties.xml")
  def_docPNG=File.join(@default, "doc_thumbnail.png")
  ###
  resXML=File.join(skm_dir, "references.xml")
  docXML=File.join(skm_dir, "document.xml")
  docPXML=File.join(skm_dir, "documentProperties.xml")
  docPNG=File.join(skm_dir, "doc_thumbnail.png")
  ### copy resXML verbatim
  file=File.open(def_resXML, "r")
    text=file.read
  file.close
  ###
  file=File.new(resXML, "w")
    file.write(text)
  file.close
  ###
  tpath=File.join(skm_dir, "doc_thumbnail.png")
  ### make inis
  fromini=File.join(@temp, "from.ini")
  toini=File.join(@temp, "to.ini")
  rini=File.join(@temp, "r.ini")
  gini=File.join(@temp, "g.ini")
  bini=File.join(@temp, "b.ini")
  ###
  ### do texture
  ###
  if mat.texture
    ###
    hasTexture="1"
    xScale=mat.texture.width.to_s
    yScale=mat.texture.height.to_s
    avgColor=mat.texture.average_color.to_i.to_s
    ### Make the texture from the material
    ref_dir=File.join(skm_dir, "ref")
    begin
		Dir.mkdir(ref_dir)
	rescue
	end
    ipath=mat.save_texture(ref_dir)
    textureFilename=File.basename(ipath)
    ipath.tr!("/","\\") if @pc
    file=File.new(fromini, "w")
      file.write(ipath)
    file.close
    tpath.tr!("/","\\") if @pc
    file=File.new(toini, "w")
      file.write(tpath)
    file.close
    jar=File.join(@dir, "Thumbnailer.jar").tr("\\","/")
	jartemp=File.join(@temp, "Thumbnailer.jar").tr("\\","/")
	file=File.open(jar, "rb")
      file.binmode
      data=file.read
	file.close
	###
	file=File.open(jartemp, "wb")
	  file.binmode
	  file.write(data)
	file.close
	###
	jar=jartemp
    ###
  else ### make thumbnail from color
    ###
    hasTexture="0"
    ###
    r=mat.color.red.to_s
    g=mat.color.green.to_s
    b=mat.color.blue.to_s
    ###
    tpath.tr!("/","\\") if @pc
    file=File.new(toini, "w")
      file.write(tpath)
    file.close
    file=File.new(rini, "w")
      file.write(r)
    file.close
    file=File.new(gini, "w")
      file.write(g)
    file.close
    file=File.new(bini, "w")
      file.write(b)
    file.close
    jar=File.join(@dir, "ThumbnailerRGB.jar").tr("\\","/")
	jartemp=File.join(@temp, "ThumbnailerRGB.jar").tr("\\","/")
	file=File.open(jar, "rb")
      file.binmode
      data=file.read
	file.close
	###
	file=File.open(jartemp, "wb")
	  file.binmode
	  file.write(data)
	file.close
	###
	jar=jartemp
    ###
  end#if
  ###
  pwd=Dir.pwd
  Dir.chdir(@temp)
    File.chmod(0777, jar) unless @pc
    if jar[0].chr=="/"
        jar="file:#{jar}"
    else
        jar="file:/#{jar}"
    end#if
    UI.openURL(jar) ###########################
  Dir.chdir(pwd)
  ###
  #return
  ### wait for thumb to be made
  wait=0.0
  until File.exist?(tpath) || wait>=2.0
    sleep(0.1)
    wait+=0.1
  end
  if ! File.exist?(tpath)
    puts "SKM.export: Could not make the SKM's 'thumbnail' image."
    File.delete(fromini) if File.exist?(fromini)
    File.delete(toini) if File.exist?(toini)
    File.delete(rini) if File.exist?(rini)
    File.delete(gini) if File.exist?(gini)
    File.delete(bini) if File.exist?(bini)
    SKM.unload(skm_dir)
    return nil
  end#if
  ###
  ### add the values
  ### docPXMP
  file=File.open(def_docPXML, "r")
    text=file.read
  file.close
  ###
  text.gsub!(/TITLE/, title)
  text.gsub!(/CREATED/, created)
  text.gsub!(/MODIFIED/, modified)
  ###
  file=File.new(docPXML, "w")
    file.write(text)
  file.close
  ### docXML
  file=File.open(def_docXML, "r")
    text=file.read
  file.close
  ###
  text.gsub!(/COLORBLUE/, colorBlue)
  text.gsub!(/COLORGREEN/, colorGreen)
  text.gsub!(/COLORRED/, colorRed)
  text.gsub!(/COLORIZETYPE/, colorizeType)
  text.gsub!(/HASTEXTURE/, hasTexture)
  text.gsub!(/NAME/, title)
  text.gsub!(/USETRANS/, useTrans)
  text.gsub!(/TRANS/, trans)
  text.gsub!(/TYPE/, type)
  ###
  if mat.texture
    text.gsub!(/AVGCOLOR/, avgColor)
    text.gsub!(/TEXTUREFILE/, textureFilename)
    text.gsub!(/XSCALE/, xScale)
    text.gsub!(/YSCALE/, yScale)
    text.gsub!(/PATH/, textureFilename)
  else
    text.gsub!(/<mat:texture(?m)(.*)mat:texture>/, "\n")
  end#if
  ###
  ### attributes - if any
  if atts
    dics=[]
    atts.each{|d|dics << d[0]}
    dics.uniq!
    dicc=dics.length.to_s
    text.gsub!(/ATTRIBS>/, "xmlns:n0=\"http://sketchup.google.com/schemas/1.0/types\">")
    attxt="    <n0:AttributeDictionaries count=\"#{dicc}\">\n"
    dics.each{|d|
      ats=[]
      atts.each{|a| ats << a if a[0]==d}
      ctr=ats.length.to_s
      attxt << "      <n0:AttributeDictionary count=\"#{ctr}\" name=\"#{d}\">\n"
      atts.each{|a|
        if a[0]==d
          k=a[1]; v=a[2]
          case v
            when Fixnum
              attxt << "        <n0:Attribute key=\"#{k}\" type=\"4\">#{v}</n0:Attribute>\n"
            when Float
              attxt << "        <n0:Attribute key=\"#{k}\" type=\"6\">#{v}</n0:Attribute>\n"
            when String
              attxt << "        <n0:Attribute key=\"#{k}\" type=\"10\">#{v}</n0:Attribute>\n"
            when TrueClass
              attxt << "        <n0:Attribute key=\"#{k}\" type=\"7\">1</n0:Attribute>\n"
            when FalseClass
              attxt << "        <n0:Attribute key=\"#{k}\" type=\"7\">0</n0:Attribute>\n"
            when Array
              attxt << "        <n0:Attribute key=\"#{k}\" type=\"11\">\n"
            when NilClass
              attxt << "        <n0:Attribute key=\"#{k}\" type=\"0\">\n"
            else
              attxt << "        <n0:Attribute key=\"#{k}\" type=\"666\">\n"
          end#case
        end#if
      }
      attxt << "      <n0:AttributeDictionary>\n"
    }
    attxt << "    </n0:AttributeDictionaries>\n"
    text.gsub!(/ATTRIBS/, attxt)
  else
    text.gsub!(/ATTRIBS>/, "")
  end#if
  ###
  ### write text to file
  file=File.new(docXML, "w")
    file.write(text)
  file.close
  ###
  ### run jar to compress folder to SKM
  if @pc
    skm_dir.tr!("/","\\")
    path.tr!("/","\\")
    File.delete(path) if File.exist?(path)
    file=File.new(fromini, "w")
      file.write(skm_dir)
    file.close
    file=File.new(toini, "w")
      file.write(path)
    file.close
    ###
    jar=File.join(@dir, "ZipDirectory.jar").tr("\\","/")
	jartemp=File.join(@temp, "ZipDirectory.jar").tr("\\","/")
	###
	file=File.open(jar, "rb")
		file.binmode
		data=file.read
	file.close
	###
	file=File.open(jartemp, "wb")
		file.binmode
		file.write(data)
	file.close
	###
	jar=jartemp
    ###
    pwd=Dir.pwd
    Dir.chdir(@temp)
      UI.openURL("file:/#{jar}") ###########################
    Dir.chdir(pwd)
    ###
  else#MAC############
    File.delete(path) if File.exist?(path)
    file=File.new(fromini, "w")
      file.write(skm_dir)
    file.close
    file=File.new(toini, "w")
      file.write(path)
    file.close
    ###
    jar=File.join(@dir, "ZipDirectory.jar").tr("\\","/")
	jartemp=File.join(@temp, "ZipDirectory.jar").tr("\\","/")
	###
	file=File.open(jar, "rb")
		file.binmode
		data=file.read
	file.close
	###
	file=File.open(jartemp, "wb")
		file.binmode
		file.write(data)
	file.close
	###
	jar=jartemp
    ###
    pwd=Dir.pwd
    Dir.chdir(@temp)
      File.chmod(0777, jar)
      UI.openURL("file:/#{jar}")
    Dir.chdir(pwd)
    ###
  end#if
  ###
  wait=0.0
  until File.exist?(path) || wait>=2.0
    sleep(0.1)### wait till thumbnail is made
    wait+=0.1
  end
  sleep(0.1)
  ### remove temp .ini files
  File.delete(fromini) if File.exist?(fromini)
  File.delete(toini) if File.exist?(toini)
  File.delete(rini) if File.exist?(rini)
  File.delete(gini) if File.exist?(gini)
  File.delete(bini) if File.exist?(bini)
  SKM.unload(skm_dir)
  ###
  if path && File.exist?(path)
    return path
  else
    puts "SKM.export: Could not make the '#{path}' file."
    return nil
  end#if

end#export


def SKM.unload(folder="")

  if folder==""
    puts "Usage: SKM.unload(folder)"
    return nil
  end#if
	###
	fodler = SKMtools.encoder(folder)
	###
  if File.exist?(folder) && File.ftype(folder)=="directory"
    begin
      Dir.entries(folder).each{|f|
        next if f=='.'
        next if f=='..'
        if f=="ref"
          Dir.entries(File.join(folder, f)).each{|ff|
            next if ff=='.'
            next if ff=='..'
            File.delete(File.join(File.join(folder, f),ff))
          }
          Dir.delete(File.join(folder, f))
        elsif
          File.delete(File.join(folder, f))
        end#if
      }
      Dir.delete(folder)
     rescue
      return nil
     end#begin
  end#if
  return true

end#unload


def SKM.save_texture(path="", folder="")

  if ! path.class==String || path==""
    puts "Usage: SKM.save_texture(path_skm, folder)"
    return nil
  end#if
	###
	path = SKMtools.encoder(path)
	###
  if ! File.exist?(path)
    puts "Usage: SKM.save_texture(path_skm, folder)"
    puts "File '#{path}' not found."
    return nil
  end#if
  if ! File.extname(path).downcase == ".skm"
    puts "Usage: SKM.save_texture(path_skm, folder)"
    puts "File '#{path}' is not a SKM file."
    return nil
  end#if
  if ! folder.class==String || folder==""
    puts "Usage: SKM.save_texture(path_skm, folder)"
    return nil
  end#if
	###
	folder = SKMtools.encoder(folder)
	###
  if not File.exist?(folder)
    puts "Usage: SKM.save_texture(path_skm, folder)"
    puts "Folder '#{folder}' was not found."
    return nil
  end#if
  if not File.ftype(folder)=="directory"
    puts "Usage: SKM.save_texture(path_skm, folder)"
    puts "Folder '#{folder}' is not a 'Directory'."
    return nil
  end#if
  ###
  oskm=File.basename(path)
  name=File.basename(path, ".*")#.gsub(/[^\.A-Za-z0-9_-]/, "_")
  ###
  skm_dir=File.join(@temp, name)
  SKM.unload(skm_dir)
  ###
  tskm=File.join(@temp, oskm)
  File.delete(tskm) if File.exist?(tskm)
  ###
  file=File.open(path, "rb")
    file.binmode
    data=file.read
  file.close
  ###
  file=File.open(tskm, "wb")
    file.binmode
    file.write(data)
  file.close
  ### make ini
  fromini=File.join(@temp, "from.ini")
  ###
  tskm.tr!("/","\\") if @pc
  file=File.new(fromini, "w")
    file.write(tskm)
  file.close
  ###
    jar=File.join(@dir, "UnZipDirectory.jar").tr("\\","/")
    jartemp=File.join(@temp, "UnZipDirectory.jar").tr("\\","/")
	###
	file=File.open(jar, "rb")
		file.binmode
		data=file.read
	file.close
	###
	file=File.open(jartemp, "wb")
		file.binmode
		file.write(data)
	file.close
	###
	jar=jartemp
  ###
  pwd=Dir.pwd
  Dir.chdir(@temp)
    File.chmod(0777, jar) unless @pc
    if jar[0].chr=="/"
        jar="file:#{jar}"
    else
        jar="file:/#{jar}"
    end#if
    UI.openURL(jar)
  Dir.chdir(pwd)
  ###
  wait=0.0
  until File.exist?(File.join(skm_dir, "document.xml")) || wait>=2.0
    sleep(0.1)### wait till xml is made
    wait+=0.1
  end#
  sleep(0.1)
  if ! File.exist?(File.join(skm_dir, "document.xml"))
    puts "SKM.save_texture: Could not extract the SKM's 'document.xml'."
    File.delete(fromini) if File.exist?(fromini)
    File.delete(tskm) if File.exist?(tskm)
    SKM.unload(skm_dir)
    return nil
  end#if
  ### get reference to image
  file=File.open(File.join(skm_dir, "document.xml"))
    xmltxt=file.read
  file.close
  if ! xmltxt.include?("<mat:texture ")
    puts "SKM.save_texture(): SKM has NO texture set in its 'document.xml'."
    File.delete(fromini) if File.exist?(fromini)
    File.delete(tskm) if File.exist?(tskm)
    SKM.unload(skm_dir)
    return nil
  end#if
  ### try to save texture 'tex'
  ref=File.join(skm_dir, "ref")
  wait=0.0
  until File.exist?(ref) || wait>=2.0
    sleep(0.1)### wait till file is made
    wait+=0.1
  end#
  sleep(0.1)
  if ! File.exist?(ref)
    tex=nil
    puts "SKM.save_texture(): SKM contains NO texture 'ref' folder."
    File.delete(fromini) if File.exist?(fromini)
    File.delete(tskm) if File.exist?(tskm)
    SKM.unload(skm_dir)
    return nil
  end#if
  tfi=nil
  Dir.entries(ref).each{|e|
    next if e=="."
    next if e==".."
    tfi=e
  }
  if ! tfi
    tex=nil
    puts "SKM.save_texture(): SKM contains NO texture file in its 'ref' folder."
    File.delete(fromini) if File.exist?(fromini)
    File.delete(tskm) if File.exist?(tskm)
    SKM.unload(skm_dir)
    return nil
  end#if
  tmpimg=File.join(ref, tfi)
  file=File.open(tmpimg, "r")
    file.binmode
    data=file.read
  file.close
  ###
  tex=File.join(folder, tfi)
  file=File.open(tex, "w")
    file.binmode
    file.write(data)
  file.close
  ###
  File.delete(fromini) if File.exist?(fromini)
  File.delete(tskm) if File.exist?(tskm)
  SKM.unload(skm_dir)
  if tex
    tex.tr!("/","\\") if @pc
    return tex
  else
    return nil
  end#if

end#save_texture


def SKM.save_thumbnail(path="", thumb="")

  if ! path.class==String || path==""
    puts "Usage: SKM.save_thumbnail(path_skm, thumb)"
    return nil
  end#if
	###
	path = SKMtools.encoder(path)
	###
  if ! File.exist?(path)
    puts "Usage: SKM.save_thumbnail(path_skm, thumb)"
    puts "File '#{path}' not found."
    return nil
  end#if
  if ! File.extname(path).downcase == ".skm"
    puts "Usage: SKM.save_thumbnail(path_skm, thumb)"
    puts "File '#{path}' is not a SKM file."
    return nil
  end#if
	###
	thumb = SKMtools.encoder(thumb)
	###
  if File.exist?(thumb) && File.ftype(thumb)=="directory"
    thumb=File.join(thumb, "doc_thumbnail.png")
  end#if
  if not File.exist?(File.dirname(thumb))
    puts "Usage: SKM.save_thumbnail(path_skm, thumb)"
    puts "Folder for '#{thumb}' was not found."
    return nil
  end#if
  if File.extname(thumb) != ".png"
    thumb=File.join(File.dirname(thumb), File.basename(thumb, ".*")) + ".png"
  end#if
  ###
  oskm=File.basename(path)
  name=File.basename(path, ".*")#.gsub(/[^\.A-Za-z0-9_-]/, "_")
  ###
  skm_dir=File.join(@temp, name)
  SKM.unload(skm_dir)
  ###
  tskm=File.join(@temp, oskm)
  File.delete(tskm) if File.exist?(tskm)
  ###
  file=File.open(path, "rb")
    file.binmode
    data=file.read
  file.close
  ###
  file=File.open(tskm, "wb")
    file.binmode
    file.write(data)
  file.close
  ### make ini
  fromini=File.join(@temp, "from.ini")
  ###
  tskm.tr!("/","\\") if @pc
  file=File.new(fromini, "w")
    file.write(tskm)
  file.close
  ###
    jar = File.join(@dir, "UnZipDirectory.jar").tr("\\","/")
    jartemp=File.join(@temp, "UnZipDirectory.jar").tr("\\","/")
	###
	file=File.open(jar, "rb")
		file.binmode
		data=file.read
	file.close
	###
	file=File.open(jartemp, "wb")
		file.binmode
		file.write(data)
	file.close
	###
	jar=jartemp
  ###
  pwd=Dir.pwd
  Dir.chdir(@temp)
    File.chmod(0777, jar) unless @pc
    if jar[0].chr=="/"
        jar="file:#{jar}"
    else
        jar="file:/#{jar}"
    end#if
    UI.openURL(jar)
  Dir.chdir(pwd)
  ###
  wait=0.0
  until File.exist?(File.join(skm_dir, "doc_thumbnail.png")) || wait>=2.0
    sleep(0.1)### wait till thumb is made
    wait+=0.1
  end#
  sleep(0.1)
  if ! File.exist?(File.join(skm_dir, "doc_thumbnail.png"))
    puts "SKM.save_thumbnail: Could not extract the SKM's 'doc_thumbnail.png'."
    File.delete(fromini) if File.exist?(fromini)
    File.delete(tskm) if File.exist?(tskm)
    SKM.unload(skm_dir)
    return nil
  end#if
  ### try to save thumbnail 'thumb'
  tmpimg=File.join(skm_dir, "doc_thumbnail.png")
  file=File.open(tmpimg, "rb")
    file.binmode
    data=file.read
  file.close
  ###
  file=File.open(thumb, "wb")
    file.binmode
    file.write(data)
  file.close
  ###
  File.delete(fromini) if File.exist?(fromini)
  File.delete(tskm) if File.exist?(tskm)
  SKM.unload(skm_dir)
  if File.exist?(thumb)
    thumb.tr!("/","\\") if @pc
    return thumb
  else
    return nil
  end#if

end#save_thumbnail


def SKM.extract(path="", folder=nil)

  if ! path.class==String || path==""
    puts "Usage: SKM.extract(path_skm, [folder])"
    return nil
  end#if
	###
	path = SKMtools.encoder(path)
	###
  if ! File.exist?(path)
    puts "Usage: SKM.extract(path_skm, [folder])"
    puts "File '#{path}' not found."
    return nil
  end#if
  if ! File.extname(path).downcase == ".skm"
    puts "Usage: SKM.extract(path_skm, [folder])"
    puts "File '#{path}' is not a SKM file."
    return nil
  end#if
  if folder
		###
		folder = SKMtools.encoder(folder)
		###
    if File.exist?(folder)
      if ! File.ftype(folder)=="directory"
        puts "Usage: SKM.extract(path_skm, folder)"
        puts "File '#{folder}' is not a 'Directory'."
        return nil
      end#if
    else
      puts "Usage: SKM.extract(path_skm, folder)"
      puts "File '#{folder}' was not found."
      return nil
    end#if
  end#if
  ###
  oskm=File.basename(path)
  name=File.basename(path, ".*").gsub(/[^\.A-Za-z0-9_-]/, "_")
  ###
  if folder
    fold=folder
  else
    fold=@temp
  end#if
  skm_dir=File.join(fold, name)
  SKM.unload(skm_dir)
  ###
  tskm=File.join(fold, oskm)
  tskm.tr!("/","\\") if @pc
  extg=false
  extg=true if tskm == path
  ###
 if ! extg
  file=File.open(path, "rb")
    file.binmode
    data=file.read
  file.close
  ###
  file=File.open(tskm, "wb")
    file.binmode
    file.write(data)
  file.close
  ### make ini
  fromini=File.join(@temp, "from.ini")
  ###
  file=File.new(fromini, "w")
    file.write(tskm)
  file.close
 end#if
  ###
  jar = File.join(@dir, "UnZipDirectory.jar").tr("\\","/")
  jartemp=File.join(@temp, "UnZipDirectory.jar").tr("\\","/")
	###
	file=File.open(jar, "rb")
		file.binmode
		data=file.read
	file.close
	###
	file=File.open(jartemp, "wb")
		file.binmode
		file.write(data)
	file.close
	###
	jar=jartemp
  ###
  pwd=Dir.pwd
  Dir.chdir(@temp)
    File.chmod(0777, jar) unless @pc
    if jar[0].chr=="/"
        jar="file:#{jar}"
    else
        jar="file:/#{jar}"
    end#if
    UI.openURL(jar)
  Dir.chdir(pwd)
  ###
  wait=0.0
  until File.exist?(File.join(skm_dir, "document.xml")) || wait>=2.0
    sleep(0.1)### wait till xml is made
    wait+=0.1
  end#
  sleep(0.1)
  if ! File.exist?(File.join(skm_dir, "document.xml"))
    puts "SKM.extract: Could not extract the SKM's 'document.xml'."
    File.delete(fromini) if File.exist?(fromini)
    File.delete(tskm) if File.exist?(tskm) && ! extg
    SKM.unload(skm_dir)
    return nil
  end#if
  ### we have the data files - read colors etc
  ### get values from xml files
  file=File.open(File.join(skm_dir, "document.xml"))
    text=file.read
  file.close
  ### parse values - colors
  r=(text.split(" colorRed=\""))[1].split("\"")[0].to_i
  g=(text.split(" colorGreen=\""))[1].split("\"")[0].to_i
  b=(text.split(" colorBlue=\""))[1].split("\"")[0].to_i
  ### alpha
  if (text.split(" useTrans=\""))[1].split("\"")[0].to_i
    a=1.0
  else
    a=(text.split(" trans=\""))[1].split("\"")[0].to_f
  end#if
  ### path to image IF set
  if text.include?(" path=\"")
    t=(text.split(" path=\""))[1].split("\"")[0]
    w=(text.split(" xScale=\""))[1].split("\"")[0].to_f
    h=(text.split(" yScale=\""))[1].split("\"")[0].to_f
    av=(text.split(" avgColor=\""))[1].split("\"")[0].to_i
  else
    t=nil
    w=nil
    h=nil
    av=nil
  end#if
  ### sort out attributes #################################################
  if text.include?("<n0:AttributeDictionaries count=\"")
    attdics_num=text.split("<n0:AttributeDictionaries count=\"")[1].split("\"")[0].to_i
    attdics_data=text.split("<n0:AttributeDictionary count=\"")[1..-1]
    attdics=[]
    attdics_data.each{|dic|
      dname=dic.split("\"")[2]
      kvs=[]
      dic.each{|d|
        d.split("<n0:Attribute key=\"")[1..-1].each{|kv|
          key=kv.split("\"")[0]
          type=kv.split(" type=\"")[1].split("\"")[0]
          if kv.include?("</n0:Attribute>")
            value=kv.split(">")[1].split("</n0:Attribute")[0]
          else
            value=nil
          end#if
          kvs << [key, type, value]
        }
      }
      attdics << [dname, kvs]
    }
    ### make attributes array
    ads=[]
    attdics.each{|ad|
      dname=ad[0]
      kvs=ad[1]
      kvs.each{|kv|
        key=kv[0]
        type=kv[1]
        value=kv[2]
        case type
          when "0"
            value=nil
          when "4"
            value=value.to_i
          when "6"
            value=value.to_f
          when "7"
            if value=="1"
              value=true
            else
              value=false
            end#if
          when "10"
            value=value
          when "11"
            value=[]
          else
            value=nil
        end#case
        ads << [dname, key, value]
      }
    }
  end#if
  ###
  ### write data to hash
  hash={}
  hash[:path]=path
  hash[:name]=name
  hash[:texture]=t
  hash[:avgcolor]=av
  hash[:xscale]=w
  hash[:yscale]=h
  hash[:color]=[r,g,b]
  hash[:alpha]=a
  hash[:attributes]=ads
  ###
  if folder
    skm_dir.tr!("/","\\")if @pc
    hash[:extracted_path]=skm_dir
    tn=File.join(skm_dir, "doc_thumbnail.png")
    tn.tr!("/","\\")if @pc
    hash[:extracted_thumbnail]=tn
    tf=nil
    tf=File.join(File.join(skm_dir, "ref"), t) if t
    tf.tr!("/","\\")if tf && @pc
    hash[:extracted_texture]=tf
    File.delete(tskm) if File.exist?(tskm) && ! extg
  else
    hash[:extracted_path]=nil
    hash[:extracted_thumbnail]=nil
    hash[:extracted_texture]=nil
    File.delete(tskm) if File.exist?(tskm) && ! extg
    SKM.unload(skm_dir)
  end#if
  ###
  File.delete(fromini) if File.exist?(fromini)
  ###
  return hash
  ###
  
end#extract


def SKM.save_texture_and_thumbnail(path="", folder=nil)
  ###
  if ! path.class==String || path==""
    puts "Usage: SKM.save_texture_and_thumbnail(path_skm, folder)"
    return nil
  end#if
	###
	path = SKMtools.encoder(path)
	###
  if ! File.exist?(path)
    puts "Usage: SKM.save_texture_and_thumbnail(path_skm, folder)"
    puts "File '#{path}' not found."
    return nil
  end#if
  if ! File.extname(path).downcase == ".skm"
    puts "Usage: SKM.save_texture_and_thumbnail(path_skm, folder)"
    puts "File '#{path}' is not a SKM file."
    return nil
  end#if
  if folder
		###
		folder = SKMtools.encoder(folder)
		###
    if File.exist?(folder)
      if ! File.ftype(folder)=="directory"
        puts "Usage: SKM.save_texture_and_thumbnail(path_skm, folder)"
        puts "File '#{folder}' is not a 'Directory'."
        return nil
      end#if
    else
      puts "Usage: SKM.save_texture_and_thumbnail(path_skm, folder)"
      puts "File '#{folder}' was not found."
      return nil
    end#if
  end#if
  ###
  oskm=File.basename(path)
  name=File.basename(path, ".*").gsub(/[^\.A-Za-z0-9_-]/, "_")
  ###
  data=SKM.extract(path, folder)
  return nil if not data
  ###
  skmdir=data[:extracted_path]
  return nil if not File.exist?(skmdir)
  ### remove all XML files
  Dir.entries(skmdir).each{|e| File.delete(File.join(skmdir, e)) if File.extname(e).downcase==".xml" }
  ### move texture if any from ref folder
  ref=File.join(skmdir, "ref")
  if File.exist?(ref) && File.ftype(ref)=="directory"
    tex=nil
    Dir.entries(ref).each{|e|
      next if e=="."
      next if e==".."
      tex=e
      break
    }
    if tex
      ### copy tex into skmdir and erase it AND res
      texf=File.join(ref, tex)
      file=File.open(texf, "rb")
        file.binmode
        dat=file.read
      file.close
      ntexf=File.join(skmdir, tex)
      file=File.open(ntexf, "wb")
        file.binmode
        file.write(dat)
      file.close
      File.delete(texf) if File.exist?(texf)
      Dir.delete(ref) if File.exist?(ref)
    end#end
  end#if
  ###
  return skmdir
  ###
end#save_texture_and_thumbnail


end#SKM
###
